"""Module for reinforcement steel material properties."""

from dataclasses import dataclass, field
from enum import Enum

from blueprints.type_alias import DIMENSIONLESS, KG_M3, MPA, PER_MILLE

REBAR_STEEL_YOUNG_MODULUS = 200_000.0  # [MPa] from EN 1992-1-1 3.2.7(4)


class ReinforcementType(Enum):
    """Enumeration of product form / reinforcement type."""

    BARS = "Bars"
    DECOILED_RODS = "Decoiled rods"
    WIRE_FABRICS = "Wire fabrics"
    LATTICE_GIRDERS = "Lattice girders"


class ReinforcementSteelQuality(Enum):
    """Enumeration of reinforcement steel quality."""

    B400A = "B400A"
    B500A = "B500A"
    B600A = "B600A"
    B400B = "B400B"
    B500B = "B500B"
    B600B = "B600B"
    B400C = "B400C"
    B500C = "B500C"
    B600C = "B600C"
    B550A = "B550A"
    B550B = "B550B"


class ReinforcementSteelClass(Enum):
    """Enumeration of reinforcement steel classes."""

    A = "A"
    B = "B"
    C = "C"


class ReinforcementBarSurface(Enum):
    """Enumeration of types of bar surface."""

    SMOOTH = "Smooth"
    RIBBED = "Ribbed"


class SteelFabrication(Enum):
    """Enumeration of types of fabrication."""

    HOT_ROLLED = "Hot-rolled"
    COLD_WORKED = "Cold-worked"


class ReinforcementDiagramType(Enum):
    """Enumeration of types of stress-strain diagrams."""

    BILINEAR_INCLINED = "Bi-linear with inclined branch"
    BILINEAR_NOT_INCLINED = "Bi-linear with horizontal branch"
    USER = "User-defined"


@dataclass(frozen=True)
class ReinforcementSteelMaterial:
    r"""Representation of the properties of reinforcement steel suitable for use with EN 1992-1-1:2004.

    Based on the analytical relations shown on table C.1 Annex C.

    Parameters
    ----------
    steel_quality: ReinforcementSteelQuality,
        Steel quality of the ReinforcementSteelMaterial object (default: B500B).
    density: KG_M3
        Unit weight of steel [kg/m³] (default= 7850.0) [$kg/m^3$]
    reinforcement_type: ReinforcementType
        Product form / Reinforcement type (default=ReinforcementType.BARS)
    bar_surface: ReinforcementBarSurface
        Type of bar surface (default=ReinforcementBarSurface.RIBBED)
    steel_fabrication: SteelFabrication
        Type of fabrication (default=SteelFabrication.HOT_ROLLED)
    diagram_type: ReinforcementDiagramType
        Type of stress-strain diagram (default=ReinforcementDiagramType.BILINEAR_INCLINED)
    material_factor: DIMENSIONLESS
        Partial safety factor [$\gamma_s$] for reinforcement steel according to EN 1992-1-1 art.2.4.2.4 (1) - Table 2.1N [$-$] [default= 1.15]
        Persistent and transient        $\gamma_s = 1.15$
        Accidental design situations    $\gamma_s = 1.0$
    custom_name: str
        User-defined name of the material (default= name of steel quality; example: 'B500B')
    custom_e_s: MPA
        User-defined Young's modulus of the material, if not provided the default value is used (default=200000) [$MPa$]

    """

    steel_quality: ReinforcementSteelQuality = field(default=ReinforcementSteelQuality.B500B)
    density: KG_M3 = field(default=7850.0, metadata={"unit": "kg/m³"})
    reinforcement_type: ReinforcementType = field(default=ReinforcementType.BARS)
    bar_surface: ReinforcementBarSurface = field(default=ReinforcementBarSurface.RIBBED)
    steel_fabrication: SteelFabrication = field(default=SteelFabrication.HOT_ROLLED)
    diagram_type: ReinforcementDiagramType = field(default=ReinforcementDiagramType.BILINEAR_NOT_INCLINED)
    material_factor: DIMENSIONLESS = field(default=1.15)
    custom_name: str | None = field(default=None, compare=False)
    custom_e_s: MPA | None = field(default=None, metadata={"unit": "MPa"})

    @property
    def name(self) -> str:
        r"""Name of the reinforcement steel material.

        Returns
        -------
        str
            Example: "B500B"
        """
        if self.custom_name:
            return self.custom_name
        return self.steel_quality.value

    @property
    def e_s(self) -> MPA:
        r"""Reinforcement steel Young's modulus [$MPa$].

        Returns
        -------
        MPA
            Example: 200000.0 (for B500B)
        """
        if self.custom_e_s:
            return self.custom_e_s
        return REBAR_STEEL_YOUNG_MODULUS

    @property
    def f_yk(self) -> MPA:
        r"""[$f_{yk}$] Characteristic yield strength of reinforcement [$MPa$].

        Returns
        -------
        MPA
            Example: 500.0 (for B500B)
        """
        return float(self.steel_quality.value[1:-1])

    @property
    def f_yd(self) -> MPA:
        r"""[$f_{yd}$] Design yield strength of reinforcement (EN 1992-1-1:2004 art.3.2.7 (2)) [$MPa$].

        Returns
        -------
        MPA
            Example: 434.78 (for B500B)
        """
        return self.f_yk / self.material_factor

    @property
    def steel_class(self) -> str:
        r"""Reinforcement class.

        Returns
        -------
        str
            Example: "B" (for B500B)
        """
        return self.steel_quality.value[-1]

    @property
    def f_tk(self) -> MPA:
        r"""[$f_{tk}$] Characteristic tensile strength of reinforcement [$MPa$].

        Returns
        -------
        MPA
            Example: 540.0 (for B500B)
        """
        return self.f_yk * self.ductility_factor_k

    @property
    def ductility_factor_k(self) -> DIMENSIONLESS:
        r"""Ductility factor k [$-$] -> ([$f_{tk}$] / [$f_{yk}$]) tabel C.1 Annex C from EN 1992-1-1:2004.

        * 1.05 for steel class A
        * 1.08 for steel class B
        * 1.15 for steel class C

        Returns
        -------
        DIMENSIONLESS
            Example: 1.08 (for B500B)
        """
        match self.steel_class.lower():
            case "a":
                return 1.05
            case "b":
                return 1.08
            case "c":
                return 1.15
            case _:
                raise ValueError(f"Unknown steel class: {self.steel_class}")

    @property
    def eps_uk(self) -> PER_MILLE:
        r"""[$\varepsilon_{uk}$] Characteristic strain of reinforcement at max. load [$‰$ (per mille)] (tabel C.1 Annex C from EN 1992-1-1:2004).

        * 250 ‰ for steel class A
        * 500 ‰ for steel class B
        * 750 ‰ for steel class C

        Returns
        -------
        PER_MILLE
            Example: 500 (for B500B)
        """
        match self.steel_class.lower():
            case "a":
                return 250
            case "b":
                return 500
            case "c":
                return 750
            case _:
                raise ValueError(f"Unknown steel class: {self.steel_class}")



"""Testing sub-formula for 3.10 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_3_materials.formula_3_10 import SubForm3Dot10FictionalCrossSection


class TestSubForm3Dot10FictionalCrossSection:
    """Validation for sub-formula for 3.10 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        a_c = 42.5  # mm²
        u = 20.3  # mm
        sub_form_3_10 = SubForm3Dot10FictionalCrossSection(a_c=a_c, u=u)

        # Expected result, manually calculated
        manually_calculated_result = 4.187192

        assert sub_form_3_10 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_a_c_is_given(self) -> None:
        """Test a negative value."""
        # Example values
        a_c = -42.5  # mm²
        u = 20.3  # mm

        with pytest.raises(ValueError):
            SubForm3Dot10FictionalCrossSection(a_c=a_c, u=u)

    def test_raise_error_when_negative_u_is_given(self) -> None:
        """Test a negative value."""
        # Example values
        a_c = 42.5  # mm²
        u = -20.3  # mm

        with pytest.raises(ValueError):
            SubForm3Dot10FictionalCrossSection(a_c=a_c, u=u)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"h_0 = 2 \cdot A_c / u = 2 \cdot 42.50 / 20.30 = 4.19",
            ),
            ("short", r"h_0 = 4.19"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        a_c = 42.5  # mm²
        u = 20.3  # mm

        # Object to test
        form_3_10_sub_latex = SubForm3Dot10FictionalCrossSection(a_c=a_c, u=u).latex()

        actual = {"complete": form_3_10_sub_latex.complete, "short": form_3_10_sub_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Test the HEB enum."""

from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.heb import HEB


class TestHEB:
    """Tests for the HEB enum."""

    def test_enum_values(self) -> None:
        """Test that enum values are correctly defined."""
        assert HEB.HEB100.value == ("HEB100", 100, 100, 6, 10, 12)
        assert HEB.HEB1000.value == ("HEB1000", 1000, 300, 19, 36, 30)

    def test_enum_membership(self) -> None:
        """Test that specific values are members of the enum."""
        assert "HEB100" in [e.value[0] for e in HEB]
        assert "HEB1000" in [e.value[0] for e in HEB]

    def test_enum_uniqueness(self) -> None:
        """Test that all enum values are unique."""
        values = [e.value for e in HEB]
        assert len(values) == len(set(values))

    def test_enum_attributes(self) -> None:
        """Test that enum attributes are correctly assigned."""
        profile = HEB.HEB100
        assert profile.alias == "HEB100"
        assert profile.total_height == 100
        assert profile.top_flange_width == 100
        assert profile.top_flange_thickness == 10
        assert profile.bottom_flange_width == 100
        assert profile.bottom_flange_thickness == 10
        assert profile.web_thickness == 6
        assert profile.top_radius == 12
        assert profile.bottom_radius == 12



"""Testing formula 9.16 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_9_detailling_and_specific_rules.formula_9_16 import (
    Form9Dot16MinimumForceOnInternalBeamLine,
)
from blueprints.validations import NegativeValueError


class TestForm9Dot16MinimumForceOnInternalBeamLine:
    """Validation for formula 9.16 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        q_3 = 20  # kN/m,
        l_1 = 4.5  # m,
        l_2 = 4  # m,
        q_4 = 70  # kN,
        form_9_16 = Form9Dot16MinimumForceOnInternalBeamLine(
            q_3=q_3,
            l_1=l_1,
            l_2=l_2,
            q_4=q_4,
        )

        # Expected result, manually calculated
        manually_calculated_result = 85

        assert form_9_16 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_evaluation_lower_limit(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        q_3 = 20  # kN/m,
        l_1 = 1.5  # m,
        l_2 = 2  # m,
        q_4 = 70  # kN,
        form_9_16 = Form9Dot16MinimumForceOnInternalBeamLine(
            q_3=q_3,
            l_1=l_1,
            l_2=l_2,
            q_4=q_4,
        )
        # Expected result, manually calculated
        manually_calculated_result = 70

        assert form_9_16 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_q_3_is_given(self) -> None:
        """Test the evaluation of the result."""
        q_3 = -20  # kN/m,
        l_1 = 1.5  # m,
        l_2 = 2  # m,
        q_4 = 70  # kN,

        with pytest.raises(NegativeValueError):
            Form9Dot16MinimumForceOnInternalBeamLine(
                q_3=q_3,
                l_1=l_1,
                l_2=l_2,
                q_4=q_4,
            )

    def test_raise_error_when_negative_l_1_is_given(self) -> None:
        """Test the evaluation of the result."""
        q_3 = 20  # kN/m,
        l_1 = -1.5  # m,
        l_2 = 2  # m,
        q_4 = 70  # kN,

        with pytest.raises(NegativeValueError):
            Form9Dot16MinimumForceOnInternalBeamLine(
                q_3=q_3,
                l_1=l_1,
                l_2=l_2,
                q_4=q_4,
            )

    def test_raise_error_when_negative_l_2_is_given(self) -> None:
        """Test the evaluation of the result."""
        q_3 = 20  # kN/m,
        l_1 = 1.5  # m,
        l_2 = -2  # m,
        q_4 = 70  # kN,

        with pytest.raises(NegativeValueError):
            Form9Dot16MinimumForceOnInternalBeamLine(
                q_3=q_3,
                l_1=l_1,
                l_2=l_2,
                q_4=q_4,
            )

    def test_raise_error_when_negative_q_4_is_given(self) -> None:
        """Test the evaluation of the result."""
        q_3 = 20  # kN/m,
        l_1 = 1.5  # m,
        l_2 = 2  # m,
        q_4 = -70  # kN,

        with pytest.raises(NegativeValueError):
            Form9Dot16MinimumForceOnInternalBeamLine(
                q_3=q_3,
                l_1=l_1,
                l_2=l_2,
                q_4=q_4,
            )

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                (r"F_{tie} = min(q_3 \cdot (l_1 + l_2) / 2, Q_4) = min(20.00 \cdot (4.50 + 4.00) / 2, 70.00) = 85.00"),
            ),
            ("short", r"F_{tie} = 85.00"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        q_3 = 20  # kN/m,
        l_1 = 4.5  # m,
        l_2 = 4  # m,
        q_4 = 70  # kN,

        # Object to test
        form_9_16_latex = Form9Dot16MinimumForceOnInternalBeamLine(
            q_3=q_3,
            l_1=l_1,
            l_2=l_2,
            q_4=q_4,
        ).latex()

        actual = {"complete": form_9_16_latex.complete, "short": form_9_16_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Testing sub-formulas for 5.1 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_5_structural_analysis.formula_5_1 import (
    SubForm5Dot1ReductionFactorLengthOrHeight,
    SubForm5Dot1ReductionFactorNumberOfMembers,
)
from blueprints.validations import LessOrEqualToZeroError


class TestSubForm5Dot1ReductionFactorLengthOrHeight:
    """Validation for sub-formula (αh) for 5.1 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        length = 8  # m

        # Object to test
        sub_form_5_1 = SubForm5Dot1ReductionFactorLengthOrHeight(length=length)

        # Expected result, manually calculated
        manually_calculated_result = 0.7071

        assert sub_form_5_1 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_l_is_given(self) -> None:
        """Test a negative value."""
        # Example values
        length = -3.5

        with pytest.raises(LessOrEqualToZeroError):
            SubForm5Dot1ReductionFactorLengthOrHeight(length=length)

    def test_raise_error_when_l_is_zero(self) -> None:
        """Test a zero value."""
        # Example values
        length = 0

        with pytest.raises(LessOrEqualToZeroError):
            SubForm5Dot1ReductionFactorLengthOrHeight(length=length)

    def test_alpha_h_is_between_two_thirds_when_l_is_high(self) -> None:
        """Test if the result is 2/3 when l is high."""
        # Example values
        length = 100  # m

        # Object to test
        sub_form_5_1 = SubForm5Dot1ReductionFactorLengthOrHeight(length=length)

        assert sub_form_5_1 == pytest.approx(expected=2 / 3, rel=1e-4)

    def test_alpha_h_is_one_when_l_is_low(self) -> None:
        """Test if the result is 1 when l is low."""
        # Example values
        length = 0.1

        # Object to test
        sub_form_5_1 = SubForm5Dot1ReductionFactorLengthOrHeight(length=length)

        assert sub_form_5_1 == pytest.approx(expected=1, rel=1e-4)


class TestSubForm5Dot1ReductionFactorNumberOfMembers:
    """Validation for sub-formula (αm) for 5.1 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        members = 5

        # Object to test
        sub_form_5_1 = SubForm5Dot1ReductionFactorNumberOfMembers(members=members)

        # Expected result, manually calculated
        manually_calculated_result = 0.774596

        assert sub_form_5_1 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_m_is_given(self) -> None:
        """Test a negative value."""
        # Example values
        members = -5

        with pytest.raises(LessOrEqualToZeroError):
            SubForm5Dot1ReductionFactorNumberOfMembers(members=members)

    def test_raise_error_when_m_is_zero(self) -> None:
        """Test a zero value."""
        # Example values
        members = 0

        with pytest.raises(LessOrEqualToZeroError):
            SubForm5Dot1ReductionFactorNumberOfMembers(members=members)



"""Testing Formula 2.1b from NEN 9997-1+C2:2017: Chapter 2: Basis of geotechnical design."""

import pytest

from blueprints.codes.eurocode.nen_9997_1_c2_2017.chapter_2_basic_of_geotechnical_design.formula_2_1_b import Form2Dot1bRepresentativeValue
from blueprints.validations import NegativeValueError


class TestForm2Dot1bRepresentativeValue:
    """Validation for formula 2.1b from NEN 9997-1+C2:2017."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # example values
        psi = 1.2  # [-]
        f_k = 100  # kN

        form_2_1_b = Form2Dot1bRepresentativeValue(psi=psi, f_k=f_k)

        # manually calculated result
        manually_calculated_result = 120

        assert form_2_1_b == pytest.approx(expected=manually_calculated_result, rel=1e-9)

    def test_raise_error_if_negative_psi(self) -> None:
        """Test that a NegativeValueError is raised when a negative value is passed for psi."""
        psi = -1
        f_k = 100

        with pytest.raises(NegativeValueError):
            Form2Dot1bRepresentativeValue(psi=psi, f_k=f_k)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"F_{rep} = \psi \cdot F_k = 1.20 \cdot 100.00 = 120.00",
            ),
            ("short", r"F_{rep} = 120.00"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        psi = 1.2  # [-]
        f_k = 100  # kN

        # Object to test
        form_2_1_b_latex = Form2Dot1bRepresentativeValue(psi=psi, f_k=f_k).latex()

        actual = {"complete": form_2_1_b_latex.complete, "short": form_2_1_b_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Testing formula 5.40b of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_5_structural_analysis.formula_5_40b import Form5Dot40bCheckLateralInstability
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm5Dot40bCheckLateralInstability:
    """Validation for formula 5.40b from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        l_0t = 6.0  # m
        b = 0.3  # m
        h = 0.5  # m

        # Object to test
        formula = Form5Dot40bCheckLateralInstability(l_0t=l_0t, b=b, h=h)

        # Expected result, manually calculated
        expected_result = True

        assert formula == expected_result

    @pytest.mark.parametrize(
        ("l_0t", "b", "h"),
        [
            (-6.0, 0.3, 0.5),  # l_0t is negative
            (6.0, -0.3, 0.5),  # b is negative
            (6.0, 0.3, -0.5),  # h is negative
            (6.0, 0.0, 0.5),  # b is zero
            (6.0, 0.3, 0.0),  # h is zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, l_0t: float, b: float, h: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form5Dot40bCheckLateralInstability(l_0t=l_0t, b=b, h=h)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"CHECK \to \left( \frac{l_{0t}}{b} \leq \frac{70}{\left( h/b \right)^{1/3}} \text{ and } \frac{h}{b} \leq 3.5 \right) \to "
                r"\left( \frac{6.000}{0.300} \leq \frac{70}{\left( 0.500/0.300 \right)^{1/3}} \text{ and } "
                r"\frac{0.500}{0.300} \leq 3.5 \right) \to OK",
            ),
            ("short", r"CHECK \to OK"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        l_0t = 6.0  # m
        b = 0.3  # m
        h = 0.5  # m

        # Object to test
        latex = Form5Dot40bCheckLateralInstability(l_0t=l_0t, b=b, h=h).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 6.5 of EN 1993-1-1:2005."""

import pytest

from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_6_ultimate_limit_state.formula_6_5 import Form6Dot5UnityCheckTensileStrength
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot5UnityCheckTensileStrength:
    """Validation for formula 6.5 from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        n_ed = 7  # kN
        n_t_rd = 10  # kN
        form = Form6Dot5UnityCheckTensileStrength(n_ed=n_ed, n_t_rd=n_t_rd)

        # Expected result, manually calculated
        expected = True

        assert form == expected

    @pytest.mark.parametrize(
        ("n_ed", "n_t_rd"),
        [
            (-7, 10),  # n_ed is negative
        ],
    )
    def test_raise_error_when_negative_n_ed_is_given(self, n_ed: float, n_t_rd: float) -> None:
        """Test a negative value for n_ed."""
        with pytest.raises(NegativeValueError):
            Form6Dot5UnityCheckTensileStrength(n_ed=n_ed, n_t_rd=n_t_rd)

    @pytest.mark.parametrize(
        ("n_ed", "n_t_rd"),
        [
            (7, 0),  # n_t_rd is zero
            (7, -10),  # n_t_rd is negative
        ],
    )
    def test_raise_error_when_negative_or_zero_n_t_rd_is_given(self, n_ed: float, n_t_rd: float) -> None:
        """Test a zero or negative value for n_t_rd."""
        with pytest.raises(LessOrEqualToZeroError):
            Form6Dot5UnityCheckTensileStrength(n_ed=n_ed, n_t_rd=n_t_rd)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"CHECK \to \left( \frac{N_{Ed}}{N_{t,Rd}} \leq 1 \right) \to "
                r"\left( \frac{7.000}{10.000} \leq 1 \right) \to OK",
            ),
            ("short", r"CHECK \to OK"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        n_ed = 7  # kN
        n_t_rd = 10  # kN

        # Object to test
        latex = Form6Dot5UnityCheckTensileStrength(n_ed=n_ed, n_t_rd=n_t_rd).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Tests for the Exposure classes
according to Table 4.1 from NEN-EN 1992-1-1:2005+A1:2015+NB:2016+A1:2020: Chapter 4 - Durability and cover to reinforcement.
"""

import pytest

from blueprints.codes.eurocode.nen_en_1992_1_1_a1_2020.chapter_4_durability_and_cover.table_4_1 import (
    Carbonation,
    Chemical,
    Chloride,
    ChlorideSeawater,
    FreezeThaw,
    Table4Dot1ExposureClasses,
)


class TestCarbonation:
    """Testing Carbonation class."""

    def test_options(self) -> None:
        """Check if the options method returns all the possible options within an exposure class."""
        assert Carbonation.options() == ["Not applicable", "XC1", "XC2", "XC3", "XC4"]

    def test_description(self) -> None:
        """Check if the description method returns the description of the subclass."""
        assert Carbonation.exposure_class_description() == "Corrosion induced by carbonation"

    def test_description_of_the_environment(self) -> None:
        """Check if the description_of_the_environment method returns the description of the environment."""
        assert Carbonation.XC1.description_of_the_environment() == "Dry or permanently wet"
        assert Carbonation.XC2.description_of_the_environment() == "Wet, rarely dry"
        assert Carbonation.XC3.description_of_the_environment() == "Moderate humidity"
        assert Carbonation.XC4.description_of_the_environment() == "Cyclic wet and dry"
        assert Carbonation.NA.description_of_the_environment() == "Not applicable"


class TestChloride:
    """Testing Chloride class."""

    def test_options(self) -> None:
        """Check if the options method returns all the possible options within an exposure class."""
        assert Chloride.options() == ["Not applicable", "XD1", "XD2", "XD3"]

    def test_description(self) -> None:
        """Check if the description method returns the description of the subclass."""
        assert Chloride.exposure_class_description() == "Corrosion induced by chlorides other than by sea water"

    def test_description_of_the_environment(self) -> None:
        """Check if the description_of_the_environment method returns the description of the environment."""
        assert Chloride.XD1.description_of_the_environment() == "Moderate humidity"
        assert Chloride.XD2.description_of_the_environment() == "Wet, rarely dry"
        assert Chloride.XD3.description_of_the_environment() == "Cyclic wet and dry"
        assert Chloride.NA.description_of_the_environment() == "Not applicable"


class TestChlorideSeawater:
    """Testing ChlorideSeawater class."""

    def test_options(self) -> None:
        """Check if the options method returns all the possible options within an exposure class."""
        assert ChlorideSeawater.options() == ["Not applicable", "XS1", "XS2", "XS3"]

    def test_description(self) -> None:
        """Check if the description method returns the description of the subclass."""
        assert ChlorideSeawater.exposure_class_description() == "Corrosion induced by chlorides from sea water"

    def test_description_of_the_environment(self) -> None:
        """Check if the description_of_the_environment method returns the description of the environment."""
        assert ChlorideSeawater.XS1.description_of_the_environment() == "Exposed to airborne salt but not in direct contact with sea water"
        assert ChlorideSeawater.XS2.description_of_the_environment() == "Permanently submerged"
        assert ChlorideSeawater.XS3.description_of_the_environment() == "Tidal, splash and spray zones"
        assert ChlorideSeawater.NA.description_of_the_environment() == "Not applicable"


class TestFreezeThaw:
    """Testing FreezeThaw class."""

    def test_options(self) -> None:
        """Check if the options method returns all the possible options within an exposure class."""
        assert FreezeThaw.options() == ["Not applicable", "XF1", "XF2", "XF3", "XF4"]

    def test_description(self) -> None:
        """Check if the description method returns the description of the subclass."""
        assert FreezeThaw.exposure_class_description() == "Freeze/thaw attack with or without de-icing agents"

    def test_description_of_the_environment(self) -> None:
        """Check if the description_of_the_environment method returns the description of the environment."""
        assert FreezeThaw.XF1.description_of_the_environment() == "Moderate water saturation, without de-icing agent"
        assert FreezeThaw.XF2.description_of_the_environment() == "Moderate water saturation, with de-icing agent"
        assert FreezeThaw.XF3.description_of_the_environment() == "High water saturation, without de-icing agents"
        assert FreezeThaw.XF4.description_of_the_environment() == "High water saturation with de-icing agents or sea water"
        assert FreezeThaw.NA.description_of_the_environment() == "Not applicable"


class TestChemical:
    """Testing Chemical class."""

    def test_options(self) -> None:
        """Check if the options method returns all the possible options within an exposure class."""
        assert Chemical.options() == ["Not applicable", "XA1", "XA2", "XA3"]

    def test_description(self) -> None:
        """Check if the description method returns the description of the subclass."""
        assert Chemical.exposure_class_description() == "Chemical attack"

    def test_description_of_the_environment(self) -> None:
        """Check if the description_of_the_environment method returns the description of the environment."""
        assert Chemical.XA1.description_of_the_environment() == "Slightly aggressive chemical environment according to EN 206-1, Table 2"
        assert Chemical.XA2.description_of_the_environment() == "Moderately aggressive chemical environment according to EN 206-1, Table 2"
        assert Chemical.XA3.description_of_the_environment() == "Highly aggressive chemical environment according to EN 206-1, Table 2"
        assert Chemical.NA.description_of_the_environment() == "Not applicable"


def test_comparing_different_types_raises_error() -> None:
    """Check if comparing different exposure class types, raises TypeError."""
    with pytest.raises(TypeError):
        _ = Carbonation.XC2 > Chloride.XD1  # type: ignore[operator]
    with pytest.raises(TypeError):
        _ = Chloride.NA == Chemical.NA  # type: ignore[operator]
    with pytest.raises(TypeError):
        _ = FreezeThaw.XF1 <= ChlorideSeawater.XS2  # type: ignore[operator]
    with pytest.raises(TypeError):
        _ = ChlorideSeawater.XS3 >= Chloride.XD3  # type: ignore[operator]
    with pytest.raises(TypeError):
        _ = Chemical.XA1 < Carbonation.XC2  # type: ignore[operator]
    with pytest.raises(TypeError):
        _ = Chemical.NA != FreezeThaw.XF1


class TestTable4Dot1ExposureClasses:
    """Unit tests for the Table4Dot1ExposureClasses class."""

    def test_table_4_1_exposure_classes_from_exposure_list(self) -> None:
        """Test the from_exposure_list method of Table4Dot1ExposureClasses."""
        exposure_classes = Table4Dot1ExposureClasses.from_exposure_list(["XC1", "XD1", "XS1", "XF1", "XA1"])
        assert exposure_classes.carbonation == Carbonation.XC1
        assert exposure_classes.chloride == Chloride.XD1
        assert exposure_classes.chloride_seawater == ChlorideSeawater.XS1
        assert exposure_classes.freeze_thaw == FreezeThaw.XF1
        assert exposure_classes.chemical == Chemical.XA1

        # test with wrong exposure class
        with pytest.raises(ValueError):
            _ = Table4Dot1ExposureClasses.from_exposure_list(["XC1", "wrong", "XS1", "XF1", "XA1"])

    def test_dubble_exposure_class(self) -> None:
        """Test the from_exposure_list method of Table4Dot1ExposureClasses with a double exposure class."""
        with pytest.raises(ValueError) as excinfo:
            _ = Table4Dot1ExposureClasses.from_exposure_list(["XC1", "XD1", "XD2", "XF1", "XA1"])

        assert str(excinfo.value) == "Duplication Error: There are multiple instances of 'Chloride' class."



"""Testing formula 7.4 from EN 1995-1-1:2004."""

from contextlib import AbstractContextManager
from contextlib import nullcontext as does_not_raise

import pytest

from blueprints.codes.eurocode.en_1995_1_1_2004.chapter_7_serviceability_limit_states.formula_7_4 import Form7Dot4VelocityResponseLimit
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm7Dot4VelocityResponseLimit:
    """Validation for formula 7.4 from EN 1995-1-1:2004."""

    @pytest.mark.parametrize(
        ("b", "f_1", "ksi", "expected_result"),
        [
            (120, 5, 0.02, 0.0134),
            (50, 2, 1, 50.000),
            (150, 3, 1.2, 454817.612),
            (120, 1, 0, 0.0083),
        ],
        ids=["b=120", "ksi = 1", "ksi > 1", "ksi = 0"],
    )
    def test_evaluation(self, b: float, f_1: float, ksi: float, expected_result: float) -> None:
        """Test the evaluation of the result."""
        form_7_4 = Form7Dot4VelocityResponseLimit(b=b, f_1=f_1, ksi=ksi)
        assert form_7_4 == pytest.approx(expected=expected_result, abs=1e-4)

    @pytest.mark.parametrize(
        ("b", "f_1", "ksi", "expectation"),
        [
            (120, 5, 0.02, does_not_raise()),
            (50, 2, -1, pytest.raises(NegativeValueError)),
            (50, 0, 1, pytest.raises(LessOrEqualToZeroError)),
            (
                40,
                8,
                1.2,
                pytest.raises(ValueError),
            ),
        ],
        ids=[
            "passes",
            "ksi < 0",
            "f_1 = 0",
            "b_outside_range",
        ],
    )
    def test_raise_error_incorrect_args(self, b: float, f_1: float, ksi: float, expectation: AbstractContextManager) -> None:
        """Test if errors are raised."""
        with expectation:
            assert Form7Dot4VelocityResponseLimit(b=b, f_1=f_1, ksi=ksi) is not None

    @pytest.mark.parametrize(
        ("b", "f_1", "ksi", "rep_short", "rep_long"),
        [
            (
                50,  # DIMENSIONLESS
                2,  # HZ
                1,  # DIMENSIONLESS
                r"v_{lim} = 50.000",
                r"v_{lim} = b^{(f_1 \cdot \xi - 1)} = 50.000^{(2.000 \cdot 1.000 - 1)} = 50.000",
            ),
            (
                120,  # DIMENSIONLESS
                5,  # HZ
                0.02,  # DIMENSIONLESS
                r"v_{lim} = 0.013",
                r"v_{lim} = b^{(f_1 \cdot \xi - 1)} = 120.000^{(5.000 \cdot 0.020 - 1)} = 0.013",
            ),
        ],
        ids=["decimal result", "integer result"],
    )
    def test_latex(self, b: float, f_1: float, ksi: float, rep_short: str, rep_long: str) -> None:
        """Test the latex representation of the formula."""
        form_7_4_latex = Form7Dot4VelocityResponseLimit(f_1=f_1, b=b, ksi=ksi).latex()
        assert form_7_4_latex.complete == rep_long
        assert form_7_4_latex.short == rep_short



"""I-Profile steel section."""

from collections.abc import Callable
from dataclasses import dataclass
from typing import Self

from matplotlib import pyplot as plt

from blueprints.materials.steel import SteelMaterial
from blueprints.structural_sections.cross_section_cornered import CircularCorneredCrossSection
from blueprints.structural_sections.cross_section_rectangle import RectangularCrossSection
from blueprints.structural_sections.steel.steel_cross_sections._steel_cross_section import CombinedSteelCrossSection
from blueprints.structural_sections.steel.steel_cross_sections.plotters.general_steel_plotter import plot_shapes
from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.hea import HEA
from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.heb import HEB
from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.hem import HEM
from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.ipe import IPE
from blueprints.structural_sections.steel.steel_element import SteelElement
from blueprints.type_alias import MM


@dataclass(kw_only=True)
class ISteelProfile(CombinedSteelCrossSection):
    """Representation of an I-Profile steel section.
    This can be used to create a custom I-profile or to create an I-profile from a standard profile.

    For standard profiles, use the `from_standard_profile` class method.
    For example,
    ```python
    i_profile = ISteelProfile.from_standard_profile(profile=HEA.HEA200, steel_material=SteelMaterial(SteelStrengthClass.S355))
    ```

    Attributes
    ----------
    steel_material : SteelMaterial
        Steel material properties for the profile.
    top_flange_width : MM
        The width of the top flange [mm].
    top_flange_thickness : MM
        The thickness of the top flange [mm].
    bottom_flange_width : MM
        The width of the bottom flange [mm].
    bottom_flange_thickness : MM
        The thickness of the bottom flange [mm].
    total_height : MM
        The total height of the profile [mm].
    web_thickness : MM
        The thickness of the web [mm].
    top_radius : MM | None
        The radius of the curved corners of the top flange. Default is None, the corner radius is then taken as the thickness.
    bottom_radius : MM | None
        The radius of the curved corners of the bottom flange. Default is None, the corner radius is then taken as the thickness.
    name : str
        The name of the profile. Default is "I-Profile". If corrosion is applied, the name will include the corrosion value.
    """

    steel_material: SteelMaterial
    top_flange_width: MM
    top_flange_thickness: MM
    bottom_flange_width: MM
    bottom_flange_thickness: MM
    total_height: MM
    web_thickness: MM
    top_radius: MM | None = None
    bottom_radius: MM | None = None
    name: str = "I-Profile"

    def __post_init__(self) -> None:
        """Initialize the I-profile steel section."""
        self.top_radius = self.top_radius if self.top_radius is not None else self.top_flange_thickness
        self.bottom_radius = self.bottom_radius if self.bottom_radius is not None else self.bottom_flange_thickness

        # Calculate web height
        self.web_height = self.total_height - self.top_flange_thickness - self.bottom_flange_thickness - self.top_radius - self.bottom_radius
        self.width_outstand_top_flange = (self.top_flange_width - self.web_thickness - 2 * self.top_radius) / 2
        self.width_outstand_bottom_flange = (self.bottom_flange_width - self.web_thickness - 2 * self.bottom_radius) / 2

        # Create the cross-sections for the flanges and web
        self.top_right_flange = RectangularCrossSection(
            name="Top Right Flange",
            width=self.width_outstand_top_flange,
            height=self.top_flange_thickness,
            x=self.top_flange_width / 2 - self.width_outstand_top_flange / 2,
            y=(self.total_height - self.top_flange_thickness) / 2,
        )

        self.top_left_flange = RectangularCrossSection(
            name="Top Left Flange",
            width=self.width_outstand_top_flange,
            height=self.top_flange_thickness,
            x=-self.top_flange_width / 2 + self.width_outstand_top_flange / 2,
            y=(self.total_height - self.top_flange_thickness) / 2,
        )

        self.bottom_right_flange = RectangularCrossSection(
            name="Bottom Right Flange",
            width=self.width_outstand_bottom_flange,
            height=self.bottom_flange_thickness,
            x=self.bottom_flange_width / 2 - self.width_outstand_bottom_flange / 2,
            y=-(self.total_height - self.bottom_flange_thickness) / 2,
        )

        self.bottom_left_flange = RectangularCrossSection(
            name="Bottom Left Flange",
            width=self.width_outstand_bottom_flange,
            height=self.bottom_flange_thickness,
            x=-self.bottom_flange_width / 2 + self.width_outstand_bottom_flange / 2,
            y=-(self.total_height - self.bottom_flange_thickness) / 2,
        )

        self.web = RectangularCrossSection(
            name="Web",
            width=self.web_thickness,
            height=self.web_height,
            x=0,
            y=(-self.top_flange_thickness - self.top_radius + self.bottom_flange_thickness + self.bottom_radius) / 2,
        )

        # Create curves for the corners of the flanges
        self.curve_top_right = CircularCorneredCrossSection(
            name="Curve top right",
            inner_radius=self.top_radius,
            outer_radius=0,
            x=self.top_radius + self.web_thickness / 2,
            y=self.total_height / 2 - self.top_flange_thickness - self.top_radius,
            corner_direction=1,
            thickness_horizontal=self.web_thickness / 2,
            thickness_vertical=self.top_flange_thickness,
        )
        self.curve_top_left = CircularCorneredCrossSection(
            name="Curve top left",
            inner_radius=self.top_radius,
            outer_radius=0,
            x=-self.top_radius - self.web_thickness / 2,
            y=self.total_height / 2 - self.top_flange_thickness - self.top_radius,
            corner_direction=0,
            thickness_horizontal=self.web_thickness / 2,
            thickness_vertical=self.top_flange_thickness,
        )
        self.curve_bottom_right = CircularCorneredCrossSection(
            name="Curve bottom right",
            inner_radius=self.bottom_radius,
            outer_radius=0,
            x=self.bottom_radius + self.web_thickness / 2,
            y=-self.total_height / 2 + self.bottom_flange_thickness + self.bottom_radius,
            corner_direction=2,
            thickness_horizontal=self.web_thickness / 2,
            thickness_vertical=self.bottom_flange_thickness,
        )
        self.curve_bottom_left = CircularCorneredCrossSection(
            name="Curve bottom left",
            inner_radius=self.bottom_radius,
            outer_radius=0,
            x=-self.bottom_radius - self.web_thickness / 2,
            y=-self.total_height / 2 + self.bottom_flange_thickness + self.bottom_radius,
            corner_direction=3,
            thickness_horizontal=self.web_thickness / 2,
            thickness_vertical=self.bottom_flange_thickness,
        )

        # Create the steel elements
        self.elements = [
            SteelElement(
                cross_section=self.top_right_flange,
                material=self.steel_material,
                nominal_thickness=self.top_flange_thickness,
            ),
            SteelElement(
                cross_section=self.top_left_flange,
                material=self.steel_material,
                nominal_thickness=self.top_flange_thickness,
            ),
            SteelElement(
                cross_section=self.bottom_right_flange,
                material=self.steel_material,
                nominal_thickness=self.bottom_flange_thickness,
            ),
            SteelElement(
                cross_section=self.bottom_left_flange,
                material=self.steel_material,
                nominal_thickness=self.bottom_flange_thickness,
            ),
            SteelElement(
                cross_section=self.web,
                material=self.steel_material,
                nominal_thickness=self.web_thickness,
            ),
            SteelElement(
                cross_section=self.curve_top_right,
                material=self.steel_material,
                nominal_thickness=self.top_flange_thickness,
            ),
            SteelElement(
                cross_section=self.curve_top_left,
                material=self.steel_material,
                nominal_thickness=self.top_flange_thickness,
            ),
            SteelElement(
                cross_section=self.curve_bottom_right,
                material=self.steel_material,
                nominal_thickness=self.bottom_flange_thickness,
            ),
            SteelElement(
                cross_section=self.curve_bottom_left,
                material=self.steel_material,
                nominal_thickness=self.bottom_flange_thickness,
            ),
        ]

    @classmethod
    def from_standard_profile(
        cls,
        profile: HEA | HEB | HEM | IPE,
        steel_material: SteelMaterial,
        corrosion: MM = 0,
    ) -> Self:
        """Create an I-profile from a set of standard profiles already defined in Blueprints.

        Blueprints offers standard profiles for HEA, HEB, HEM, and IPE. This method allows you to create an I-profile.

        Parameters
        ----------
        profile : HEA | HEB | HEM | IPE
            Any of the standard profiles defined in Blueprints.
        steel_material : SteelMaterial
            Steel material properties for the profile.
        corrosion : MM, optional
            Corrosion thickness per side (default is 0).
        """
        top_flange_width = profile.top_flange_width - corrosion * 2
        top_flange_thickness = profile.top_flange_thickness - corrosion * 2
        bottom_flange_width = profile.bottom_flange_width - corrosion * 2
        bottom_flange_thickness = profile.bottom_flange_thickness - corrosion * 2
        total_height = profile.total_height - corrosion * 2
        web_thickness = profile.web_thickness - corrosion * 2

        if any(
            [
                top_flange_thickness < 1e-3,
                bottom_flange_thickness < 1e-3,
                web_thickness < 1e-3,
            ]
        ):
            raise ValueError("The profile has fully corroded.")

        name = profile.alias
        if corrosion:
            name += f" (corrosion: {corrosion} mm)"

        return cls(
            top_flange_width=top_flange_width,
            top_flange_thickness=top_flange_thickness,
            bottom_flange_width=bottom_flange_width,
            bottom_flange_thickness=bottom_flange_thickness,
            total_height=total_height,
            web_thickness=web_thickness,
            steel_material=steel_material,
            top_radius=profile.top_radius,
            bottom_radius=profile.bottom_radius,
            name=name,
        )

    def plot(self, plotter: Callable[[CombinedSteelCrossSection], plt.Figure] | None = None, *args, **kwargs) -> plt.Figure:
        """Plot the cross-section. Making use of the standard plotter.

        Parameters
        ----------
        plotter : Callable[CombinedSteelCrossSection, plt.Figure] | None
            The plotter function to use. If None, the default Blueprints plotter for steel sections is used.
        *args
            Additional arguments passed to the plotter.
        **kwargs
            Additional keyword arguments passed to the plotter.
        """
        if plotter is None:
            plotter = plot_shapes
        return plotter(
            self,
            *args,
            **kwargs,
        )



"""Formula 6.65 from EN 1992-1-1:2004: Chapter 6 - Ultimate Limit State."""

import numpy as np

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DEG
from blueprints.validations import raise_if_negative


class Form6Dot65ConcreteCompressionStrut(Formula):
    r"""Class representing formula 6.65 for the calculation of [$\theta_{fat}$]."""

    label = "6.65"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        theta: DEG,
    ) -> None:
        r"""[$\theta_{fat}$] Calculation of concrete compression struts angle for fatigue.

        EN 1992-1-1:2004 art.6.8.2(2) - Formula (6.65)

        Parameters
        ----------
        theta : DEG
            [$\theta$] Angle of concrete compression struts to the beam axis assumed in ULS design [$degrees$].
        """
        super().__init__()
        self.theta = theta

    @staticmethod
    def _evaluate(
        theta: DEG,
    ) -> DEG:
        """Evaluates the formula, for more information see the __init__ method."""
        theta_rad = np.deg2rad(theta)
        tangent_theta = np.tan(theta_rad)
        raise_if_negative(tangent_theta=tangent_theta)

        theta_fat_rad = np.arctan(np.minimum(np.sqrt(np.tan(theta_rad)), 1))
        return np.rad2deg(theta_fat_rad)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.65."""
        _equation: str = r"\tan^{-1}\left(\min\left(\sqrt{\tan(\theta)}, 1\right)\right)"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"\theta": f"{self.theta:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"\theta_{fat}",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="degrees",
        )



"""Formula 7.19 from EN 1992-1-1:2004: Chapter 7 - Serviceability Limit State."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, MPA
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form7Dot19DistributionCoefficient(Formula):
    r"""Class representing formula 7.19 for the calculation of [$\zeta$]."""

    label = "7.19"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        beta: DIMENSIONLESS,
        sigma_sr: MPA,
        sigma_s: MPA,
    ) -> None:
        r"""[$\zeta$] Calculation of the distribution coefficient, 0 for uncracked sections [$\zeta$].

        Note: $\sigma_{sr} / \sigma_{s}$ may be replaced by $M_{cr} / M$ for flexure or $N_{cr} / N$ for pure tension,
        where $M_{cr}$ is the cracking moment and $N_{cr}$ is the cracking force.

        EN 1992-1-1:2004 art.7.4.3(3) - Formula (7.19)

        Parameters
        ----------
        beta : DIMENSIONLESS
            [$\beta$] Coefficient taking account of the influence of the duration of the loading or of
            repeated loading on the average strain. For short-term loading, [$\beta$] = 1.0. For sustained loads or
            many cycles of repeated loading [$\beta$] = 0.5 [$-$].
        sigma_sr : MPA
            [$\sigma_{sr}$] Stress in the tension reinforcement calculated on the basis of a
            cracked section under the loading conditions causing first cracking [$MPa$].
        sigma_s : MPA
            [$\sigma_{s}$] Stress in the tension reinforcement calculated on the basis of a cracked section under
            loading conditions causing first cracking [$MPa$].
        """
        super().__init__()
        self.beta = beta
        self.sigma_sr = sigma_sr
        self.sigma_s = sigma_s

    @staticmethod
    def _evaluate(
        beta: DIMENSIONLESS,
        sigma_sr: MPA,
        sigma_s: MPA,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(beta=beta, sigma_sr=sigma_sr, sigma_s=sigma_s)
        raise_if_less_or_equal_to_zero(sigma_s=sigma_s)

        return 1 - beta * (sigma_sr / sigma_s) ** 2

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 7.19."""
        _equation: str = r"1 - \beta \left(\frac{\sigma_{sr}}{\sigma_{s}}\right)^2"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"\beta": f"{self.beta:.{n}f}",
                r"\sigma_{sr}": f"{self.sigma_sr:.{n}f}",
                r"\sigma_{s}": f"{self.sigma_s:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"\zeta",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="-",
        )



"""Testing formula 5.19 of EN 1993-5:2007."""

import pytest

from blueprints.codes.eurocode.en_1993_5_2007.chapter_5_ultimate_limit_states.formula_5_19 import Form5Dot19CompressionCheckClass3Profiles
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm5Dot19CompressionCheckClass3Profiles:
    """Validation for formula 5.19 from EN 1993-5:2007."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        n_ed = 9.0  # kN
        n_pl_rd = 100.0  # kN

        # Object to test
        formula = Form5Dot19CompressionCheckClass3Profiles(
            n_ed=n_ed,
            n_pl_rd=n_pl_rd,
        )

        # Expected result, manually calculated
        expected_result = True

        assert formula == expected_result
        assert formula.unity_check == n_ed / n_pl_rd

    @pytest.mark.parametrize(
        ("n_ed", "n_pl_rd"),
        [
            (10.0, 0.0),  # n_pl_rd zero
            (10.0, -80.0),  # n_pl_rd negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, n_ed: float, n_pl_rd: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form5Dot19CompressionCheckClass3Profiles(n_ed=n_ed, n_pl_rd=n_pl_rd)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"CHECK \to \frac{N_{Ed}}{N_{pl,Rd}} \leq 0.1 \to \frac{9.000}{100.000} \leq 0.1 \to OK",
            ),
            ("short", r"CHECK \to OK"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        n_ed = 9.0  # kN
        n_pl_rd = 100.0  # kN

        # Object to test
        latex = Form5Dot19CompressionCheckClass3Profiles(n_ed=n_ed, n_pl_rd=n_pl_rd).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Module with the representation of the covers for cross-sections."""

from collections import defaultdict
from dataclasses import asdict, dataclass

from blueprints.type_alias import MM
from blueprints.validations import raise_if_negative

DEFAULT_COVER = 50  # mm


@dataclass(frozen=True)
class CoversRectangular:
    """Representation of the covers of a rectangular cross-section."""

    upper: MM = DEFAULT_COVER
    right: MM = DEFAULT_COVER
    lower: MM = DEFAULT_COVER
    left: MM = DEFAULT_COVER

    def __post_init__(self) -> None:
        """Post initialization of the covers."""
        self.validate()

    def get_covers_info(self) -> str:
        """Return a string with the covers of the cross-section."""
        caption_text = "Cover:"

        covers = defaultdict(list)

        for key, value in asdict(self).items():
            covers[value].append(key)

        if len(covers) == 1:
            return f"{caption_text} {self.upper:.0f} mm"

        cover_texts = [caption_text]

        for cover, names in covers.items():
            cover_texts.append(f"{'|'.join(names)}: {cover:.0f} mm")

        return "\n  ".join(cover_texts)

    def validate(self) -> None:
        """Validate the covers."""
        raise_if_negative(upper=self.upper, right=self.right, lower=self.lower, left=self.left)



"""Formula 9.7N from EN 1992-1-1:2004: Chapter 9 - Detailing of members and particular rules."""

import numpy as np

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DEG, MM
from blueprints.validations import raise_if_greater_than_90, raise_if_negative


class Form9Dot7nMaximumDistanceBentUpBars(Formula):
    """Class representing the formula 9.7N for the calculation of the maximum distance between bent-up bars in longitudinal direction."""

    label = "9.7N"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        d: MM,
        alpha: DEG,
    ) -> None:
        r"""[$s_{b,max}$] Maximum distance between bent-up bars in longitudinal direction [$mm$].

        EN 1992-1-1:2004 art.9.2.2(7) - Formula (9.7N)

        Parameters
        ----------
        d: MM
            [$d$] Effective height of the cross-section [$mm$].
        alpha: DEG
            [$\alpha$] The angle between the shear reinforcement and the longitudinal axis of the beam (see 9.2.2(1)) [$deg$].
        """
        super().__init__()
        self.d = d
        self.alpha = alpha

    @staticmethod
    def _evaluate(d: MM, alpha: DEG) -> MM:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(d=d, alpha=alpha)
        raise_if_greater_than_90(alpha=alpha)

        # Convert the angle from degrees to radians
        alpha_radians = np.deg2rad(alpha)

        # Calculate the cotangent
        cot_alpha = 1 / np.tan(alpha_radians)

        return 0.6 * d * (1 + cot_alpha)

    def latex(self, n: int = 2) -> LatexFormula:
        """Returns LatexFormula object for formula 9.7N."""
        return LatexFormula(
            return_symbol=r"s_{b,max}",
            result=f"{self:.{n}f}",
            equation=r"0.6 \cdot d \cdot \left( 1 + cot(\alpha) \right)",
            numeric_equation=rf"0.6 \cdot {self.d:.{n}f} \cdot \left( 1 + cot({self.alpha:.{n}f}) \right)",
            comparison_operator_label="=",
        )



"""Test suite for StripSteelProfile."""

from unittest.mock import MagicMock

import matplotlib.pyplot as plt
import pytest

from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_3_materials.table_3_1 import SteelStrengthClass
from blueprints.materials.steel import SteelMaterial
from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.strip import Strip
from blueprints.structural_sections.steel.steel_cross_sections.strip_profile import StripSteelProfile


class TestStripSteelProfile:
    """Test suite for StripSteelProfile."""

    def test_code(self, strip_profile: StripSteelProfile) -> None:
        """Test the code of the Strip profile."""
        expected_alias = "160x5"
        assert strip_profile.name == expected_alias

    def test_steel_volume_per_meter(self, strip_profile: StripSteelProfile) -> None:
        """Test the steel volume per meter."""
        expected_volume = 0.160 * 0.005  # m³/m
        assert pytest.approx(strip_profile.volume_per_meter, rel=1e-6) == expected_volume

    def test_steel_weight_per_meter(self, strip_profile: StripSteelProfile) -> None:
        """Test the steel weight per meter."""
        expected_weight = 0.160 * 0.005 * 7850  # kg/m
        assert pytest.approx(strip_profile.weight_per_meter, rel=1e-6) == expected_weight

    def test_area(self, strip_profile: StripSteelProfile) -> None:
        """Test the steel cross-sectional area."""
        expected_area = 160 * 5  # mm²
        assert pytest.approx(strip_profile.area, rel=1e-6) == expected_area

    @pytest.mark.slow
    def test_plot(self, strip_profile: StripSteelProfile) -> None:
        """Test the plot method (ensure it runs without errors)."""
        fig = strip_profile.plot(show=False)
        assert isinstance(fig, plt.Figure)

    def test_plot_mocked(self, strip_profile: StripSteelProfile, mock_section_properties: MagicMock) -> None:  # noqa: ARG002
        """Test the plotting of the Strip profile shapes with mocked section properties."""
        fig = strip_profile.plot(show=False)
        assert isinstance(fig, plt.Figure)

    def test_geometry(self, strip_profile: StripSteelProfile) -> None:
        """Test the geometry of the Strip profile."""
        expected_geometry = strip_profile.geometry
        assert expected_geometry is not None

    def test_yield_strength(self, strip_profile: StripSteelProfile) -> None:
        """Test the yield strength of the Strip profile."""
        assert strip_profile.yield_strength == 355

    def test_ultimate_strength(self, strip_profile: StripSteelProfile) -> None:
        """Test the ultimate strength of the Strip profile."""
        assert strip_profile.ultimate_strength == 490

    def test_get_profile_with_corrosion(self) -> None:
        """Test the Strip profile with 2 mm corrosion applied."""
        # Ensure the profile raises an error if fully corroded
        with pytest.raises(ValueError, match="The profile has fully corroded."):
            StripSteelProfile.from_standard_profile(
                profile=Strip.STRIP160x5,
                steel_material=SteelMaterial(SteelStrengthClass.S355),
                corrosion=2.5,
            )

    def test_corrosion_in_name(self, strip_profile: StripSteelProfile) -> None:
        """Test that the corrosion is included in the profile name."""
        profile_with_corrosion = StripSteelProfile.from_standard_profile(
            profile=Strip.STRIP160x5,
            steel_material=SteelMaterial(SteelStrengthClass.S355),
            corrosion=1,
        )
        expected_name = f"{strip_profile.name} (corrosion: 1 mm)"
        assert profile_with_corrosion.name == expected_name



"""Testing Formula 2.4 from NEN 9997-1+C2:2017: Chapter 2: Basis of geotechnical design."""

import pytest

from blueprints.codes.eurocode.nen_9997_1_c2_2017.chapter_2_basic_of_geotechnical_design.formula_2_4 import Form2Dot4DesignValueGeotechnicalParameter
from blueprints.validations import NegativeValueError


class TestForm2Dot4DesignValueGeotechnicalParameter:
    """Validation for formula 2.4 from NEN 9997-1+C2:2017."""

    def test_evaluation_lower_than(self) -> None:
        """Test the evaluation of the result."""
        assert Form2Dot4DesignValueGeotechnicalParameter(e_dst_d=0.2, e_stb_d=0.31, t_d=-0.1)

    def test_evaluation_equals(self) -> None:
        """Test the evaluation of the result."""
        assert Form2Dot4DesignValueGeotechnicalParameter(e_dst_d=0.2, e_stb_d=0.3, t_d=-0.1)

    def test_evaluation_greater_than(self) -> None:
        """Test the evaluation of the result."""
        assert not Form2Dot4DesignValueGeotechnicalParameter(e_dst_d=2.01, e_stb_d=1, t_d=1)

    def test_raise_error_if_negative_e_dst_d(self) -> None:
        """Test that a NegativeValueError is raised when a negative value is passed for E_dst;d."""
        with pytest.raises(NegativeValueError):
            bool(Form2Dot4DesignValueGeotechnicalParameter(e_dst_d=-1, e_stb_d=1, t_d=1))

    def test_raise_error_if_negative_e_stb_d(self) -> None:
        """Test that a NegativeValueError is raised when a negative value is passed for E_stb;d."""
        with pytest.raises(NegativeValueError):
            bool(Form2Dot4DesignValueGeotechnicalParameter(e_dst_d=1, e_stb_d=-1, t_d=1))

    def test_str(self) -> None:
        """Test the string representation of the formula."""
        assert isinstance(str(Form2Dot4DesignValueGeotechnicalParameter(e_dst_d=2.01, e_stb_d=1, t_d=1)), str)

    @pytest.mark.parametrize(
        ("representation", "expected_result"),
        [
            ("complete", r"E_{dst;d} \leq E_{stb;d} + T_d \to 2.01 \leq 1.00 + 1.00 \to 2.01 \leq 2.00 \to \text{Not OK}"),
            ("short", r" \to \text{Not OK}"),
        ],
    )
    def test_latex(self, representation: str, expected_result: str) -> None:
        """Test the latex representation of the formula."""
        # Object to test
        latex = Form2Dot4DesignValueGeotechnicalParameter(e_dst_d=2.01, e_stb_d=1, t_d=1).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert actual[representation] == expected_result, f"{representation} representation failed."



"""Testing formula 6.25 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_25 import Form6Dot25DesignShearResistance
from blueprints.validations import GreaterThan90Error, LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot25DesignShearResistance:
    """Validation for formula 6.25 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        c = 0.5
        mu = 0.6
        f_ctd = 2.0
        sigma_n = 1.0
        a_s = 1000.0
        a_i = 200000.0
        f_yd = 500.0
        alpha = 30.0
        nu = 0.9
        f_cd = 30.0

        # Object to test
        formula = Form6Dot25DesignShearResistance(
            c=c, mu=mu, f_ctd=f_ctd, sigma_n=sigma_n, a_s=a_s, a_i=a_i, f_yd=f_yd, alpha=alpha, nu=nu, f_cd=f_cd
        )

        # Expected result, manually calculated
        manually_calculated_result = 4.51506350946  # MPa

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("c", "mu", "f_ctd", "sigma_n", "a_s", "a_i", "f_yd", "alpha", "nu", "f_cd"),
        [
            (-0.5, 0.6, 2.0, 1.0, 1000.0, 200000.0, 500.0, 30.0, 0.9, 30.0),  # c is negative
            (0.5, -0.6, 2.0, 1.0, 1000.0, 200000.0, 500.0, 30.0, 0.9, 30.0),  # mu is negative
            (0.5, 0.6, -2.0, 1.0, 1000.0, 200000.0, 500.0, 30.0, 0.9, 30.0),  # f_ctd is negative
            (0.5, 0.6, 2.0, -1.0, 1000.0, 200000.0, 500.0, 30.0, 0.9, 30.0),  # sigma_n is negative
            (0.5, 0.6, 2.0, 1.0, -1000.0, 200000.0, 500.0, 30.0, 0.9, 30.0),  # a_s is negative
            (0.5, 0.6, 2.0, 1.0, 1000.0, -200000.0, 500.0, 30.0, 0.9, 30.0),  # a_i is negative
            (0.5, 0.6, 2.0, 1.0, 1000.0, 200000.0, -500.0, 30.0, 0.9, 30.0),  # f_yd is negative
            (0.5, 0.6, 2.0, 1.0, 1000.0, 200000.0, 500.0, -30.0, 0.9, 30.0),  # alpha is negative
            (0.5, 0.6, 2.0, 1.0, 1000.0, 200000.0, 500.0, 30.0, -0.9, 30.0),  # nu is negative
            (0.5, 0.6, 2.0, 1.0, 1000.0, 200000.0, 500.0, 30.0, 0.9, -30.0),  # f_cd is negative
            (0.5, 0.6, 2.0, 1.0, 1000.0, 0.0, 500.0, 30.0, 0.9, 30.0),  # a_i is zero
            (0.5, 0.6, 2.0, 1.0, 1000.0, 200000.0, 500.0, 100.0, 0.9, 30.0),  # alpha is greater than 90
        ],
    )
    def test_raise_error_when_invalid_values_are_given(
        self, c: float, mu: float, f_ctd: float, sigma_n: float, a_s: float, a_i: float, f_yd: float, alpha: float, nu: float, f_cd: float
    ) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError, GreaterThan90Error)):
            Form6Dot25DesignShearResistance(c=c, mu=mu, f_ctd=f_ctd, sigma_n=sigma_n, a_s=a_s, a_i=a_i, f_yd=f_yd, alpha=alpha, nu=nu, f_cd=f_cd)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"v_{Rdi} = \min \left( c \cdot f_{ctd} + \mu \cdot \sigma_{n} + \frac{A_{s}}{A_{i}} \cdot "
                r"f_{yd} \cdot (\mu \cdot \sin(\alpha) + \cos(\alpha)); 0.5 \cdot \nu \cdot f_{cd} \right) = "
                r"\min \left( 0.500 \cdot 2.000 + 0.600 \cdot 1.000 + \frac{1000.000}{200000.000} \cdot 500.000 \cdot "
                r"(0.600 \cdot \sin(30.000) + \cos(30.000)); 0.5 \cdot 0.900 \cdot 30.000 \right) = 4.515 \ MPa",
            ),
            ("short", r"v_{Rdi} = 4.515 \ MPa"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        c = 0.5
        mu = 0.6
        f_ctd = 2.0
        sigma_n = 1.0
        a_s = 1000.0
        a_i = 200000.0
        f_yd = 500.0
        alpha = 30.0
        nu = 0.9
        f_cd = 30.0

        # Object to test
        latex = Form6Dot25DesignShearResistance(
            c=c, mu=mu, f_ctd=f_ctd, sigma_n=sigma_n, a_s=a_s, a_i=a_i, f_yd=f_yd, alpha=alpha, nu=nu, f_cd=f_cd
        ).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 5.42 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_5_structural_analysis.formula_5_42 import Form5Dot42ConcreteCompressiveStress
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm5Dot42ConcreteCompressiveStress:
    """Validation for formula 5.42 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        sigma_c = 10.0  # MPa
        f_ck_t = 30.0  # MPa

        # Object to test
        formula = Form5Dot42ConcreteCompressiveStress(sigma_c=sigma_c, f_ck_t=f_ck_t)

        # Expected result, manually calculated
        expected_result = True

        assert formula == expected_result

    @pytest.mark.parametrize(
        ("sigma_c", "f_ck_t"),
        [
            (-10.0, 30.0),  # sigma_c is negative
            (10.0, -30.0),  # f_ck_t is negative
            (0.0, 30.0),  # sigma_c is zero
            (10.0, 0.0),  # f_ck_t is zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, sigma_c: float, f_ck_t: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form5Dot42ConcreteCompressiveStress(sigma_c=sigma_c, f_ck_t=f_ck_t)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"CHECK \to \sigma_{c} \leq 0.6 \cdot f_{ck}(t) \to 10.000 \leq 0.6 \cdot 30.000 \to OK",
            ),
            ("short", r"CHECK \to OK"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        sigma_c = 10.0  # MPa
        f_ck_t = 30.0  # MPa

        # Object to test
        latex = Form5Dot42ConcreteCompressiveStress(sigma_c=sigma_c, f_ck_t=f_ck_t).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Formula 6.14 from EN 1992-1-1:2004: Chapter 6 - Ultimate limit state."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.math_helpers import cot
from blueprints.type_alias import DEG, DIMENSIONLESS, MM, MPA, N
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot14MaxShearResistanceInclinedReinforcement(Formula):
    r"""Class representing formula 6.14 for the calculation of the maximum shear resistance for members with inclined
    shear reinforcement, [$V_{Rd,max}$].
    """

    label = "6.14"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        alpha_cw: DIMENSIONLESS,
        b_w: MM,
        z: MM,
        nu_1: DIMENSIONLESS,
        f_cd: MPA,
        theta: DEG,
        alpha: DEG,
    ) -> None:
        r"""[$V_{Rd,max}$] Maximum shear resistance for members with inclined shear reinforcement [$N$].

        EN 1992-1-1:2004 art.6.2.3(4) - Formula (6.14)

        Parameters
        ----------
        alpha_cw : DIMENSIONLESS
            [$\alpha_{cw}$] Coefficient taking account of the state of the stress in the compression chord [$-$].
        b_w : MM
            [$b_{w}$] Width of the web [$mm$].
        z : MM
            [$z$] Lever arm [$mm$].
        nu_1 : DIMENSIONLESS
            [$\nu_{1}$] Strength reduction factor for concrete [$-$].
        f_cd : MPA
            [$f_{cd}$] Design value of concrete compressive strength [$MPa$].
        theta : DEG
            [$\theta$] Angle between the concrete compression strut and the beam axis perpendicular to the
            shear force [$degrees$].
        alpha : DEG
            [$\alpha$] Angle between shear reinforcement and the beam axis perpendicular to the shear force [$degrees$].
        """
        super().__init__()
        self.alpha_cw = alpha_cw
        self.b_w = b_w
        self.z = z
        self.nu_1 = nu_1
        self.f_cd = f_cd
        self.theta = theta
        self.alpha = alpha

    @staticmethod
    def _evaluate(
        alpha_cw: DIMENSIONLESS,
        b_w: MM,
        z: MM,
        nu_1: DIMENSIONLESS,
        f_cd: MPA,
        theta: DEG,
        alpha: DEG,
    ) -> N:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            alpha_cw=alpha_cw,
            b_w=b_w,
            z=z,
            nu_1=nu_1,
            f_cd=f_cd,
            theta=theta,
            alpha=alpha,
        )
        denominator = 1 + cot(theta) ** 2
        raise_if_less_or_equal_to_zero(denominator=denominator)

        return alpha_cw * b_w * z * nu_1 * f_cd * (cot(theta) + cot(alpha)) / (1 + cot(theta) ** 2)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.14."""
        return LatexFormula(
            return_symbol=r"V_{Rd,max}",
            result=f"{self:.{n}f}",
            equation=r"\alpha_{cw} \cdot b_{w} \cdot z \cdot \nu_{1} \cdot f_{cd} \cdot \frac{\cot(\theta) + \cot(\alpha)}{1 + \cot^2(\theta)}",
            numeric_equation=rf"{self.alpha_cw:.{n}f} \cdot {self.b_w:.{n}f} \cdot {self.z:.{n}f} \cdot {self.nu_1:.{n}f} \cdot {self.f_cd:.{n}f} \cdot "  # noqa: E501
            rf"\frac{{\cot({self.theta:.{n}f}) + \cot({self.alpha:.{n}f})}}{{1 + \cot^2({self.theta:.{n}f})}}",
            comparison_operator_label="=",
            unit="N",
        )



"""Formula 6.13 from EN 1992-1-1:2004: Chapter 6 - Ultimate limit state."""

import numpy as np

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.math_helpers import cot
from blueprints.type_alias import DEG, MM, MM2, MPA, N
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot13ShearResistanceInclinedReinforcement(Formula):
    r"""Class representing formula 6.13 for the calculation of the shear resistance for members with inclined
    shear reinforcement, [$V_{Rd,s}$].
    """

    label = "6.13"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        a_sw: MM2,
        s: MM,
        z: MM,
        f_ywd: MPA,
        theta: DEG,
        alpha: DEG,
    ) -> None:
        r"""[$V_{Rd,s}$] Shear resistance for members with inclined shear reinforcement [$N$].

        EN 1992-1-1:2004 art.6.2.3(4) - Formula (6.13)

        Parameters
        ----------
        a_sw : MM2
            [$A_{sw}$] Area of shear reinforcement [$mm^2$].
        s : MM
            [$s$] Spacing of the shear reinforcement [$mm$].
        z : MM
            [$z$] Lever arm [$mm$].
        f_ywd : MPA
            [$f_{ywd}$] Design yield strength of the shear reinforcement [$MPa$].
        theta : DEG
            [$\theta$] angle between the concrete compression strut and the beam axis perpendicular to the
            shear force [$degrees$].
        alpha : DEG
            [$\alpha$] angle between shear reinforcement and the beam axis perpendicular to the shear force [$degrees$].
        """
        super().__init__()
        self.a_sw = a_sw
        self.s = s
        self.z = z
        self.f_ywd = f_ywd
        self.theta = theta
        self.alpha = alpha

    @staticmethod
    def _evaluate(
        a_sw: MM2,
        s: MM,
        z: MM,
        f_ywd: MPA,
        theta: DEG,
        alpha: DEG,
    ) -> N:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            a_sw=a_sw,
            z=z,
            f_ywd=f_ywd,
            theta=theta,
            alpha=alpha,
        )
        raise_if_less_or_equal_to_zero(
            s=s,
        )

        return (a_sw / s) * z * f_ywd * (cot(theta) + cot(alpha)) * np.sin(np.deg2rad(alpha))

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.13."""
        return LatexFormula(
            return_symbol=r"V_{Rd,s}",
            result=f"{self:.{n}f}",
            equation=r"\frac{A_{sw}}{s} \cdot z \cdot f_{ywd} \cdot \left(\cot(\theta) + \cot(\alpha)\right) \cdot \sin(\alpha)",
            numeric_equation=rf"\frac{{{self.a_sw:.{n}f}}}{{{self.s:.{n}f}}} \cdot {self.z:.{n}f} \cdot {self.f_ywd:.{n}f} \cdot "
            rf"\left(\cot({self.theta:.{n}f}) + \cot({self.alpha:.{n}f})\right) \cdot \sin({self.alpha:.{n}f})",
            comparison_operator_label="=",
            unit="N",
        )



"""Formula 7.5 from EN 1995-1-1:2004."""

import math

from blueprints.codes.eurocode.en_1995_1_1_2004 import EN_1995_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import HZ, KG_M2, NM2_M, M
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form7Dot5NaturalFrequency(Formula):
    r"""Class representing formula 7.5 for the calculation of natural frequency [$f_{1}$]."""

    label = "7.5"
    source_document = EN_1995_1_1_2004

    def __init__(self, length: M, ei_l: NM2_M, m: KG_M2) -> None:
        r"""[$f_{1}$] The natural frequency [$Hz$].

        EN 1995-1-1:2004 art 7.3.3(4) - Formula (7.5)

        Parameters
        ----------
        length : M
            [$l$] Span of the floor [$m$].
        ei_l : NM2_M
            [$(EI)_{l}$] Equivalent plate bending stiffness of the floor around the axis perpendicular to the
            longitudinal axis of the beam [$Nm^{2}/m$].
        m : KG_M2
            [$m$] Mass per unit area [$kg/m^{2}$].

        Returns
        -------
        None
        """
        super().__init__()
        self.length = length
        self.ei_l = ei_l
        self.m = m

    @staticmethod
    def _evaluate(length: M, ei_l: NM2_M, m: KG_M2) -> HZ:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(m=m, length=length)
        raise_if_negative(ei_l=ei_l)
        sqrt_ei_m = (ei_l / m) ** 0.5
        return (math.pi / (2 * length**2)) * sqrt_ei_m

    def latex(self, n: int = 2) -> LatexFormula:  # noqa: ARG002
        """Returns LatexFormula object for formula 7.5."""
        eq1: str = r"\frac{\pi}{2 \cdot l^{2}} \cdot \sqrt{\frac{(EI)_{l}}{m}}"
        repl_symb = {"m": f"{self.m:.{2}f}", " l": f" {self.length:.{2}f}", "(EI)_{l}": f"{self.ei_l:.{2}f}"}
        return LatexFormula(
            return_symbol=r"f_{1}",
            result=f"{self:.{2}f}",
            equation=eq1,
            numeric_equation=latex_replace_symbols(eq1, repl_symb),
            comparison_operator_label="=",
        )



"""Testing formula 8.10 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_8_detailing_of_reinforcement_and_prestressing_tendons.formula_8_3 import (
    Form8Dot3RequiredAnchorageLength,
)
from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_8_detailing_of_reinforcement_and_prestressing_tendons.formula_8_10 import (
    Form8Dot10DesignLapLength,
    SubForm8Dot10Alpha6,
)
from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_8_detailing_of_reinforcement_and_prestressing_tendons.formula_8_11 import (
    Form8Dot11MinimumDesignLapLength,
)
from blueprints.validations import NegativeValueError


class TestForm8Dot10DesignLapLength:
    """Validation for formula 8.10 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # example values
        alpha_1 = 1  # [-]
        alpha_2 = 1  # [-]
        alpha_3 = 1  # [-]
        alpha_5 = 1  # [-]
        alpha_6 = 1  # [-]
        l_b_rqd = 450  # mm
        l_0_min = 200  # mm
        form_8_10 = Form8Dot10DesignLapLength(
            alpha_1=alpha_1,
            alpha_2=alpha_2,
            alpha_3=alpha_3,
            alpha_5=alpha_5,
            alpha_6=alpha_6,
            l_b_rqd=l_b_rqd,
            l_0_min=l_0_min,
        )

        # manually calculated result
        manually_calculated_result = 450  # mm

        assert form_8_10 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_evaluation_minimum(self) -> None:
        """Test the evaluation of the result if the minimum is reached."""
        # example values
        alpha_1 = 1  # [-]
        alpha_2 = 1  # [-]
        alpha_3 = 1  # [-]
        alpha_5 = 1  # [-]
        alpha_6 = 1  # [-]
        l_b_rqd = 450  # mm
        l_0_min = 500  # mm
        form_8_10 = Form8Dot10DesignLapLength(
            alpha_1=alpha_1,
            alpha_2=alpha_2,
            alpha_3=alpha_3,
            alpha_5=alpha_5,
            alpha_6=alpha_6,
            l_b_rqd=l_b_rqd,
            l_0_min=l_0_min,
        )

        # manually calculated result
        manually_calculated_result = 500  # mm

        assert form_8_10 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_if_alpha_1_is_negative(self) -> None:
        """Test that a NegativeValueError is raised if alpha_1 is negative."""
        # example values
        alpha_1 = -1  # [-]
        alpha_2 = 1  # [-]
        alpha_3 = 1  # [-]
        alpha_5 = 1  # [-]
        alpha_6 = 1  # [-]
        l_b_rqd = 450  # mm
        l_0_min = 200  # mm

        with pytest.raises(NegativeValueError):
            Form8Dot10DesignLapLength(
                alpha_1=alpha_1,
                alpha_2=alpha_2,
                alpha_3=alpha_3,
                alpha_5=alpha_5,
                alpha_6=alpha_6,
                l_b_rqd=l_b_rqd,
                l_0_min=l_0_min,
            )

    def test_raise_error_if_alpha_2_is_negative(self) -> None:
        """Test that a NegativeValueError is raised if alpha_2 is negative."""
        # example values
        alpha_1 = 1  # [-]
        alpha_2 = -1  # [-]
        alpha_3 = 1  # [-]
        alpha_5 = 1  # [-]
        alpha_6 = 1  # [-]
        l_b_rqd = 450  # mm
        l_0_min = 200  # mm

        with pytest.raises(NegativeValueError):
            Form8Dot10DesignLapLength(
                alpha_1=alpha_1,
                alpha_2=alpha_2,
                alpha_3=alpha_3,
                alpha_5=alpha_5,
                alpha_6=alpha_6,
                l_b_rqd=l_b_rqd,
                l_0_min=l_0_min,
            )

    def test_raise_error_if_alpha_3_is_negative(self) -> None:
        """Test that a NegativeValueError is raised if alpha_3 is negative."""
        # example values
        alpha_1 = 1  # [-]
        alpha_2 = 1  # [-]
        alpha_3 = -1  # [-]
        alpha_5 = 1  # [-]
        alpha_6 = 1  # [-]
        l_b_rqd = 450  # mm
        l_0_min = 200  # mm

        with pytest.raises(NegativeValueError):
            Form8Dot10DesignLapLength(
                alpha_1=alpha_1,
                alpha_2=alpha_2,
                alpha_3=alpha_3,
                alpha_5=alpha_5,
                alpha_6=alpha_6,
                l_b_rqd=l_b_rqd,
                l_0_min=l_0_min,
            )

    def test_raise_error_if_alpha_5_is_negative(self) -> None:
        """Test that a NegativeValueError is raised if alpha_5 is negative."""
        alpha_1 = 1  # [-]
        alpha_2 = 1  # [-]
        alpha_3 = 1  # [-]
        alpha_5 = -1  # [-]
        alpha_6 = 1  # [-]
        l_b_rqd = 450  # mm
        l_0_min = 200  # mm

        with pytest.raises(NegativeValueError):
            Form8Dot10DesignLapLength(
                alpha_1=alpha_1,
                alpha_2=alpha_2,
                alpha_3=alpha_3,
                alpha_5=alpha_5,
                alpha_6=alpha_6,
                l_b_rqd=l_b_rqd,
                l_0_min=l_0_min,
            )

    def test_raise_error_if_alpha_6_is_negative(self) -> None:
        """Test that a NegativeValueError is raised if alpha_6 is negative."""
        # example values
        alpha_1 = 1  # [-]
        alpha_2 = 1  # [-]
        alpha_3 = 1  # [-]
        alpha_5 = 1  # [-]
        alpha_6 = -1  # [-]
        l_b_rqd = 450  # mm
        l_0_min = 200  # mm

        with pytest.raises(NegativeValueError):
            Form8Dot10DesignLapLength(
                alpha_1=alpha_1,
                alpha_2=alpha_2,
                alpha_3=alpha_3,
                alpha_5=alpha_5,
                alpha_6=alpha_6,
                l_b_rqd=l_b_rqd,
                l_0_min=l_0_min,
            )

    def test_raise_error_if_l_b_rqd_is_negative(self) -> None:
        """Test that a NegativeValueError is raised if l_b_rqd is negative."""
        # example values
        alpha_1 = 1  # [-]
        alpha_2 = 1  # [-]
        alpha_3 = 1  # [-]
        alpha_5 = 1  # [-]
        alpha_6 = 1  # [-]
        l_b_rqd = -450  # mm
        l_0_min = 200  # mm

        with pytest.raises(NegativeValueError):
            Form8Dot10DesignLapLength(
                alpha_1=alpha_1,
                alpha_2=alpha_2,
                alpha_3=alpha_3,
                alpha_5=alpha_5,
                alpha_6=alpha_6,
                l_b_rqd=l_b_rqd,
                l_0_min=l_0_min,
            )

    def test_raise_error_if_l_0_min_is_negative(self) -> None:
        """Test that a NegativeValueError is raised if l_0_min is negative."""
        # example values
        alpha_1 = 1  # [-]
        alpha_2 = 1  # [-]
        alpha_3 = 1  # [-]
        alpha_5 = 1  # [-]
        alpha_6 = 1  # [-]
        l_b_rqd = 450  # mm
        l_0_min = -200  # mm

        with pytest.raises(NegativeValueError):
            Form8Dot10DesignLapLength(
                alpha_1=alpha_1,
                alpha_2=alpha_2,
                alpha_3=alpha_3,
                alpha_5=alpha_5,
                alpha_6=alpha_6,
                l_b_rqd=l_b_rqd,
                l_0_min=l_0_min,
            )

    def test_integration_with_form_8_3(self) -> None:
        """Test the integration with formula 8.3."""
        # example values
        alpha_1 = 1  # [-]
        alpha_2 = 1  # [-]
        alpha_3 = 1  # [-]
        alpha_5 = 1  # [-]
        alpha_6 = 1  # [-]
        l_0_min = 200  # mm
        diameter = 16  # mm
        sigma_sd = 500  # MPa
        f_bd = 2.5  # MPa
        l_b_rqd = Form8Dot3RequiredAnchorageLength(
            diameter=diameter,
            sigma_sd=sigma_sd,
            f_bd=f_bd,
        )

        # manually calculated result
        manually_calculated_result = 800  # mm

        assert Form8Dot10DesignLapLength(
            alpha_1=alpha_1,
            alpha_2=alpha_2,
            alpha_3=alpha_3,
            alpha_5=alpha_5,
            alpha_6=alpha_6,
            l_b_rqd=l_b_rqd,
            l_0_min=l_0_min,
        ) == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_integration_with_sub_form_8_10(self) -> None:
        """Test the integration with sub formula 8.10."""
        # example values
        alpha_1 = 1  # [-]
        alpha_2 = 1  # [-]
        alpha_3 = 1  # [-]
        alpha_5 = 1  # [-]
        l_b_rqd = 450  # mm
        l_0_min = 200  # mm
        rho_1 = 0.5  # [-]
        alpha_6 = SubForm8Dot10Alpha6(
            rho_1=rho_1,
        )
        form_8_10 = Form8Dot10DesignLapLength(
            alpha_1=alpha_1,
            alpha_2=alpha_2,
            alpha_3=alpha_3,
            alpha_5=alpha_5,
            alpha_6=alpha_6,
            l_b_rqd=l_b_rqd,
            l_0_min=l_0_min,
        )

        # manually calculated result
        manually_calculated_result = 450  # mm

        assert form_8_10 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_integration_with_form_8_11(self) -> None:
        """Test the evaluation of the result."""
        # example values
        alpha_1 = 1  # [-]
        alpha_2 = 1  # [-]
        alpha_3 = 1  # [-]
        alpha_5 = 1  # [-]
        alpha_6 = 1  # [-]
        l_b_rqd = 200  # mm
        diameter = 16  # mm
        l_0_min = Form8Dot11MinimumDesignLapLength(
            alpha_6=alpha_6,
            l_b_rqd=l_b_rqd,
            diameter=diameter,
        )
        form_8_10 = Form8Dot10DesignLapLength(
            alpha_1=alpha_1,
            alpha_2=alpha_2,
            alpha_3=alpha_3,
            alpha_5=alpha_5,
            alpha_6=alpha_6,
            l_b_rqd=l_b_rqd,
            l_0_min=l_0_min,
        )

        # manually calculated result
        manually_calculated_result = 240  # mm

        assert form_8_10 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("representation", "expected_result"),
        [
            (
                "complete",
                r"l_{0} = \max \left\{\alpha_1 \cdot \alpha_2 \cdot \alpha_3 \cdot \alpha_5 \cdot \alpha_6 \cdot l_{b,rqd};"
                r" l_{0,min}\right\} = \max \left\{1.00 \cdot 1.00 \cdot 1.00 \cdot 1.00 \cdot 1.00 \cdot 200.00; 400.00\right\} = 400.00",
            ),
            ("short", "l_{0} = 400.00"),
        ],
    )
    def test_latex(self, representation: str, expected_result: str) -> None:
        """Test latex representation of the formula."""
        alpha_1 = 1  # [-]
        alpha_2 = 1  # [-]
        alpha_3 = 1  # [-]
        alpha_5 = 1  # [-]
        alpha_6 = 1  # [-]
        l_b_rqd = 200  # mm
        l_0_min = 400  # mm

        latex = Form8Dot10DesignLapLength(
            alpha_1=alpha_1,
            alpha_2=alpha_2,
            alpha_3=alpha_3,
            alpha_5=alpha_5,
            alpha_6=alpha_6,
            l_b_rqd=l_b_rqd,
            l_0_min=l_0_min,
        ).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert actual[representation] == expected_result, f"{representation} representation failed."



"""Testing formula 6.19 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_19 import Form6Dot19CheckShearForce
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot19CheckShearForce:
    """Validation for formula 6.19 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        v_ed = 100000  # N
        a_sw = 200.0  # mm²
        f_ywd = 400.0  # MPa
        alpha = 45.0  # degrees

        # Object to test
        formula = Form6Dot19CheckShearForce(v_ed=v_ed, a_sw=a_sw, f_ywd=f_ywd, alpha=alpha)

        # Expected result, manually calculated
        expected_result = False

        assert formula == expected_result

    @pytest.mark.parametrize(
        ("v_ed", "a_sw", "f_ywd", "alpha"),
        [
            (-100.0, 200.0, 400.0, 45.0),  # v_ed is negative
            (100.0, -200.0, 400.0, 45.0),  # a_sw is negative
            (100.0, 200.0, -400.0, 45.0),  # f_ywd is negative
            (100.0, 200.0, 400.0, -45.0),  # alpha is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, v_ed: float, a_sw: float, f_ywd: float, alpha: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot19CheckShearForce(v_ed=v_ed, a_sw=a_sw, f_ywd=f_ywd, alpha=alpha)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"CHECK \to V_{Ed} \leq A_{sw} \cdot f_{ywd} \cdot \sin(\alpha) \to"
                r" 100000.000 \leq 200.000 \cdot 400.000 \cdot \sin(45.000) \to \text{Not OK}",
            ),
            ("short", r"CHECK \to \text{Not OK}"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        v_ed = 100000  # N
        a_sw = 200.0  # mm²
        f_ywd = 400.0  # MPa
        alpha = 45.0  # degrees

        # Object to test
        latex = Form6Dot19CheckShearForce(v_ed=v_ed, a_sw=a_sw, f_ywd=f_ywd, alpha=alpha).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Formula 5.101 from EN 1992-2:2005: Chapter 5 - Structural Analysis."""

import numpy as np

from blueprints.codes.eurocode.en_1992_2_2005 import EN_1992_2_2005
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, M
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form5Dot101Imperfections(Formula):
    r"""Class representing formula 5.101 for the calculation of initial inclination imperfections, [$$\Theta_i$$]."""

    label = "5.101"
    source_document = EN_1992_2_2005

    def __init__(
        self,
        theta_0: DIMENSIONLESS,
        alpha_h: DIMENSIONLESS,
    ) -> None:
        r"""[$$\Theta_i$$] Initial inclination imperfections, [$$\Theta_i$$], is a ratio between height
        and inclination of the member [$$-$$].

        EN 1992-2:2005 art.5.2(5) - Formula (5.101)

        Parameters
        ----------
        theta_0 : DIMENSIONLESS
            [$$\Theta_0$$] Basic value [$$-$$].
            Note: The value of [$$\Theta_0$$] for use in a Country may be found in its National Annex.
            The recommended value is 1/200 [$$-$$].
        alpha_h : DIMENSIONLESS
            [$$\alpha_h$$] Reduction factor for length or height [$$-$$].
            Use your own implementation of this value or use the SubForm5Dot101ReductionFactorLengthOrHeight class.
        """
        super().__init__()
        self.theta_0 = theta_0
        self.alpha_h = alpha_h

    @staticmethod
    def _evaluate(
        theta_0: DIMENSIONLESS,
        alpha_h: DIMENSIONLESS,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(theta_0=theta_0, alpha_h=alpha_h)
        return theta_0 * alpha_h

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 5.101."""
        return LatexFormula(
            return_symbol=r"\theta_i",
            result=f"{self:.{n + 1}f}",
            equation=r"\theta_0 \cdot \alpha_h",
            numeric_equation=rf"{self.theta_0:.{n}f} \cdot {self.alpha_h:.{n}f}",
            comparison_operator_label="=",
        )


class Form5Dot101Sub1ReductionFactorLengthOrHeight(Formula):
    r"""Class representing sub-formula 5.101 for the calculation of the reduction factor for length
    or height, [$$\alpha_h$$].
    """

    label = "5.101"
    source_document = EN_1992_2_2005

    def __init__(
        self,
        length: M,
    ) -> None:
        r"""[$$\alpha_h$$] Reduction factor for length or height [$$-$$].

        The calculated value of [$$\alpha_h$$] is below 1.0.

        NEN-EN 1992-2+C2:2005 art.5.2(105) - Formula (5.101)

        Parameters
        ----------
        length : M
            [$$length$$] Length or height [$$m$$].
        """
        super().__init__()
        self.length = length

    @staticmethod
    def _evaluate(
        length: M,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(length=length)
        return min(1, 2 / np.sqrt(length))

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 5.101 subformula 1."""
        return LatexFormula(
            return_symbol=r"\alpha_h",
            result=f"{self:.{n}f}",
            equation=r"\min(2 / \sqrt{l}, 1)",
            numeric_equation=rf"\min( 2 / \sqrt{{{self.length:.{n}f}}}, 1)",
            comparison_operator_label="=",
        )



"""Formula 4.2 from EN 1993-1-8:2005: Chapter 4 - Welded Connections."""

from blueprints.codes.eurocode.en_1993_1_8_2005 import EN_1993_1_8_2005
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import N
from blueprints.validations import raise_if_negative


class Form4Dot2CheckWeldedConnection(Formula):
    r"""Class representing formula 4.2 for checking welded connections."""

    label = "4.2"
    source_document = EN_1993_1_8_2005

    def __init__(
        self,
        fw_ed: N,
        fw_rd: N,
    ) -> None:
        r"""Check the force in the weld per unit length against its resistance.

        EN 1993-1-8:2005 art.4.5.3.3(1) - Formula (4.2)

        Parameters
        ----------
        fw_ed : N
            [$F_{w,Ed}$] Design value of the force in the weld per unit length [$N$].
        fw_rd : N
            [$F_{w,Rd}$] Design value of the resistance of the weld per unit length [$N$].
        """
        super().__init__()
        self.fw_ed = fw_ed
        self.fw_rd = fw_rd

    @staticmethod
    def _evaluate(
        fw_ed: N,
        fw_rd: N,
    ) -> bool:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(fw_ed=fw_ed, fw_rd=fw_rd)

        return fw_ed <= fw_rd

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 4.2."""
        _equation: str = r"F_{w,Ed} \leq F_{w,Rd}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                "F_{w,Ed}": f"{self.fw_ed:.{n}f}",
                "F_{w,Rd}": f"{self.fw_rd:.{n}f}",
            },
            False,
        )
        _numeric_equation_with_units: str = latex_replace_symbols(
            _equation,
            {
                "F_{w,Ed}": rf"{self.fw_ed:.{n}f} \ N",
                "F_{w,Rd}": rf"{self.fw_rd:.{n}f} \ N",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"CHECK",
            result="OK" if self.__bool__() else "\\text{Not OK}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            numeric_equation_with_units=_numeric_equation_with_units,
            comparison_operator_label="\\to",
            unit="",
        )



"""Module for the concrete structural class
according to Table 4.3 from NEN-EN 1992-1-1:2005+A1:2015+NB:2016+A1:2020: Chapter 4 - Durability and cover to reinforcement.
"""

from collections.abc import Sequence
from typing import Self

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_4_durability_and_cover._base_classes.structural_class import (
    AbstractConcreteStructuralClassCalculator,
    ConcreteStructuralClassBase,
)
from blueprints.codes.eurocode.nen_en_1992_1_1_a1_2020 import NEN_EN_1992_1_1_A1_2020
from blueprints.codes.eurocode.nen_en_1992_1_1_a1_2020.chapter_4_durability_and_cover.table_4_1 import (
    Carbonation,
    Chloride,
    ChlorideSeawater,
    Table4Dot1ExposureClasses,
)
from blueprints.materials.concrete import ConcreteMaterial, ConcreteStrengthClass
from blueprints.type_alias import YEARS

"""Design working life in years as defined in table 4.3 of NEN-EN 1992-1-1:2005+A1:2015+NB:2016+A1:2020."""
DESIGN_WORKING_LIFE_DEFAULT = 50
DESIGN_WORKING_LIFE_75 = 75
DESIGN_WORKING_LIFE_100 = 100


class ConcreteStructuralClassCalculator(AbstractConcreteStructuralClassCalculator):
    """Implementation of the structural class calculator of the concrete element.

    In accordance with:
    NEN-EN 1992-1-1:2005+A1:2015+NB:2016+A1:2020 Concrete - General
    """

    source_document = NEN_EN_1992_1_1_A1_2020
    DEFAULT_STRUCTURAL_CLASS = 4
    DEFAULT_EXPLANATION = "Default structural class (S4)"

    def __init__(
        self,
        exposure_classes: Table4Dot1ExposureClasses | Sequence[str],
        design_working_life: YEARS,
        concrete_material: ConcreteMaterial,
        plate_geometry: bool,
        quality_control: bool,
    ) -> None:
        """Initializer of the ConcreteStructuralClassCalculator class.

        Parameters
        ----------
        exposure_classes : ExposureClasses | Sequence[str]
            The exposure classes of the concrete element. This can be a sequence of strings or an instance of the ExposureClasses class.
        design_working_life : YEARS, optional
            The design working life of the concrete element
        concrete_material : ConcreteMaterial, optional
            The concrete material of the concrete element
        plate_geometry : bool, optional
            True if the concrete element has a plate geometry, False otherwise
        quality_control : bool, optional
            True if the quality control of the concrete element is ensured, False otherwise
        """
        if not isinstance(exposure_classes, Table4Dot1ExposureClasses):
            exposure_classes = Table4Dot1ExposureClasses.from_exposure_list(exposure_classes)
        super().__init__(exposure_classes, design_working_life, concrete_material, plate_geometry, quality_control)

    def _structural_class_delta_design_working_life(self) -> None:
        """Calculates the addition to the structural class based on the design working life.

        In accordance with:
        NEN-EN 1992-1-1:2005+A1:2015+NB:2016+A1:2020 Concrete - General
        """
        if self.design_working_life > DESIGN_WORKING_LIFE_75:
            explanation = (
                f"{DESIGN_WORKING_LIFE_100} years"
                if self.design_working_life == DESIGN_WORKING_LIFE_100
                else f"{self.design_working_life} > {DESIGN_WORKING_LIFE_75} => {DESIGN_WORKING_LIFE_100} years"
            )
            self.update_structural_class(2, explanation)
        elif self.design_working_life <= DESIGN_WORKING_LIFE_DEFAULT:
            explanation = (
                f"{DESIGN_WORKING_LIFE_DEFAULT} years"
                if self.design_working_life == DESIGN_WORKING_LIFE_DEFAULT
                else f"{self.design_working_life} < {DESIGN_WORKING_LIFE_DEFAULT} => {DESIGN_WORKING_LIFE_DEFAULT} years"
            )
            self.update_structural_class(0, explanation)
        else:
            explanation = (
                f"{DESIGN_WORKING_LIFE_75} years"
                if self.design_working_life == DESIGN_WORKING_LIFE_75
                else f"{DESIGN_WORKING_LIFE_DEFAULT} < {self.design_working_life} < {DESIGN_WORKING_LIFE_75} => {DESIGN_WORKING_LIFE_75} years"
            )
            self.update_structural_class(1, explanation)

    def _structural_class_delta_quality_control(self) -> None:
        """Calculates the addition to the structural class based on the quality control.

        In accordance with:
        NEN-EN 1992-1-1:2005+A1:2015+NB:2016+A1:2020 Concrete - General
        """
        if self.quality_control:
            self.update_structural_class(-1, "quality control")
        else:
            self.update_structural_class(0, "no quality control")

    def _structural_class_delta_plate_geometry(self) -> None:
        """Calculates the addition to the structural class based on the plate geometry.

        In accordance with:
        NEN-EN 1992-1-1:2005+A1:2015+NB:2016+A1:2020 Concrete - General
        """
        if self.plate_geometry:
            self.update_structural_class(-1, "plate geometry")
        else:
            self.update_structural_class(0, "no plate geometry")

    def _structural_class_delta_concrete_grade(self) -> None:
        """Calculates the addition to the structural class based on the concrete grade.

        In accordance with:
        NEN-EN 1992-1-1:2005+A1:2015+NB:2016+A1:2020 Concrete - General
        """
        decisive_exposure_classes = {
            ConcreteStrengthClass("C45/55"): {Chloride.XD3.value, ChlorideSeawater.XS2.value, ChlorideSeawater.XS3.value},
            ConcreteStrengthClass("C40/50"): {Carbonation.XC4.value, Chloride.XD1.value, Chloride.XD2.value, ChlorideSeawater.XS1.value},
            ConcreteStrengthClass("C35/45"): {Carbonation.XC2.value, Carbonation.XC3.value},
            ConcreteStrengthClass("C30/37"): {Carbonation.XC1.value},
        }
        exposure_classes = {exposure_class.value for exposure_class in self.exposure_classes}

        for concrete_grade, decisive_exposure_class in decisive_exposure_classes.items():
            if decisive_exposure_class.intersection(exposure_classes):
                return (
                    self.update_structural_class(0, "no reduction with respect to concrete grade")
                    if self.concrete_material.f_ck < ConcreteMaterial(concrete_grade).f_ck
                    else self.update_structural_class(-1, f"concrete grade >= {concrete_grade.value}")
                )
        return (
            self.update_structural_class(0, "no reduction with respect to concrete grade")
            if self.concrete_material.f_ck < ConcreteMaterial(ConcreteStrengthClass("C30/37")).f_ck
            else self.update_structural_class(-1, "concrete grade >= C30/37")
        )


class Table4Dot3ConcreteStructuralClass(ConcreteStructuralClassBase):
    """Implementation of table 4.3 from NEN-EN 1992-1-1:2005+A1:2015+NB:2016+A1:2020 Concrete - General.

    Structural class of the concrete element.
    """

    def __new__(
        cls,
        *args,
        **kwargs,
    ) -> Self:
        """Constructor of the StructuralClass class.

        Parameters
        ----------
            value (int): The value of the structural class

        Returns
        -------
            Self (ConcreteStructuralClass): The instance of the ConcreteStructuralClass class
        """
        return super().__new__(cls, ConcreteStructuralClassCalculator, *args, **kwargs)

    def __init__(
        self,
        exposure_classes: Table4Dot1ExposureClasses | Sequence[str],
        design_working_life: YEARS,
        concrete_material: ConcreteMaterial,
        plate_geometry: bool,
        quality_control: bool,
    ) -> None:
        """Initializer of the ConcreteStructuralClass class.

        Parameters
        ----------
        exposure_classes : Table4Dot1ExposureClasses | Sequence[str]
            The exposure classes of the concrete element. This can be a sequence of strings or an instance of the ExposureClasses class.
        design_working_life : YEARS
            The design working life of the concrete element
        concrete_material : ConcreteMaterial
            The concrete material of the concrete element
        plate_geometry : bool
            True if the concrete element has a plate geometry, False otherwise
        quality_control : bool
            True if the quality control of the concrete element is ensured, False otherwise
        """



"""Testing formula 3.13 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_3_materials.formula_3_13 import Form3Dot13CoefficientTimeAutogeneShrinkage


class TestForm3Dot13CoefficientTimeAutogeneShrinkage:
    """Validation for formula 3.13 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        t = 5  # days
        form_3_13 = Form3Dot13CoefficientTimeAutogeneShrinkage(t=t)

        # Expected result, manually calculated
        manually_calculated_result = 0.3605927

        assert form_3_13 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_t_is_given(self) -> None:
        """Test formula raising error by a negative value."""
        # Example values
        t = -5  # days

        with pytest.raises(ValueError):
            Form3Dot13CoefficientTimeAutogeneShrinkage(t=t)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\beta_{as}(t) = 1 - \exp(-0.2 \cdot t^{0.5}) = 1 - \exp(-0.2 \cdot 5.00^{0.5}) = 0.36",
            ),
            ("short", r"\beta_{as}(t) = 0.36"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        t = 5  # days

        # Object to test
        form_3_13_latex = Form3Dot13CoefficientTimeAutogeneShrinkage(t=t).latex()

        actual = {"complete": form_3_13_latex.complete, "short": form_3_13_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Testing formula 7.8 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_7_serviceability_limit_state.formula_7_8 import Form7Dot8CrackWidth
from blueprints.validations import NegativeValueError


class TestForm7Dot8CrackWidth:
    """Validation for formula 7.8 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        s_r_max = 200.0
        epsilon_sm_minus_epsilon_cm = 0.001

        # Object to test
        formula = Form7Dot8CrackWidth(s_r_max=s_r_max, epsilon_sm_minus_epsilon_cm=epsilon_sm_minus_epsilon_cm)

        # Expected result, manually calculated
        manually_calculated_result = 0.200  # mm

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("s_r_max", "epsilon_sm_minus_epsilon_cm"),
        [
            (-200.0, 0.001),  # s_r_max is negative
            (200.0, -0.001),  # epsilon_sm_minus_epsilon_cm is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, s_r_max: float, epsilon_sm_minus_epsilon_cm: float) -> None:
        """Test invalid values."""
        with pytest.raises(NegativeValueError):
            Form7Dot8CrackWidth(s_r_max=s_r_max, epsilon_sm_minus_epsilon_cm=epsilon_sm_minus_epsilon_cm)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"w_k = s_{r,max} \cdot (\epsilon_{sm} - \epsilon_{cm}) = 200.000 \cdot (0.001) = 0.200 \ mm",
            ),
            ("short", r"w_k = 0.200 \ mm"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        s_r_max = 200.0
        epsilon_sm_minus_epsilon_cm = 0.001

        # Object to test
        latex = Form7Dot8CrackWidth(s_r_max=s_r_max, epsilon_sm_minus_epsilon_cm=epsilon_sm_minus_epsilon_cm).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Tests for hexagonal cross-section shapes."""

import numpy as np
import pytest

from blueprints.structural_sections.cross_section_hexagon import HexagonalCrossSection


class TestHexagonalCrossSection:
    """Tests for the HexagonalCrossSection class."""

    def test_area(self, hexagonal_cross_section: HexagonalCrossSection) -> None:
        """Test the area property of the HexagonalCrossSection class."""
        expected_area = (3 * np.sqrt(3) / 2) * 50.0**2
        assert hexagonal_cross_section.area == pytest.approx(expected=expected_area, rel=1e-6)

    def test_polygon(self, hexagonal_cross_section: HexagonalCrossSection) -> None:
        """Test the geometry property of the HexagonalCrossSection class."""
        polygon = hexagonal_cross_section.polygon
        assert polygon.is_valid
        assert polygon.area == pytest.approx(expected=hexagonal_cross_section.area, rel=1e-4)

    def test_invalid_side_length(self) -> None:
        """Test initialization with an invalid side length value."""
        with pytest.raises(ValueError, match="Side length must be a positive value"):
            HexagonalCrossSection(name="InvalidHexagon", side_length=-10.0, x=0.0, y=0.0)

    def test_geometry(self, hexagonal_cross_section: HexagonalCrossSection) -> None:
        """Test the geometry property of the HexagonalCrossSection class."""
        geometry = hexagonal_cross_section.geometry()
        assert geometry is not None

    def test_apothem(self, hexagonal_cross_section: HexagonalCrossSection) -> None:
        """Test the apothem property of the HexagonalCrossSection class."""
        expected_apothem = hexagonal_cross_section.side_length * np.sqrt(3) / 2
        assert hexagonal_cross_section.apothem == pytest.approx(expected=expected_apothem, rel=1e-6)

    def test_perimter(self, hexagonal_cross_section: HexagonalCrossSection) -> None:
        """Test the perimeter property of the HexagonalCrossSection class."""
        expected_perimeter = 6 * hexagonal_cross_section.side_length
        assert hexagonal_cross_section.perimeter == pytest.approx(expected=expected_perimeter, rel=1e-6)

    def test_parameters_as_dict(self, hexagonal_cross_section: HexagonalCrossSection) -> None:
        """Test the parameters_as_dict method of the HexagonalCrossSection class."""
        params = hexagonal_cross_section.section_properties().asdict()
        assert params



"""Testing math helpers for Blueprints."""

import pytest

from blueprints.math_helpers import cot, csc, sec
from blueprints.validations import GreaterThan90Error, LessOrEqualToZeroError, NegativeValueError


class TestCot:
    """Validation for cotangent function."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        x = 45.0
        expected_result = 1.0
        assert cot(x) == pytest.approx(expected_result, rel=1e-4)

    @pytest.mark.parametrize("x", [0.0, -45.0])
    def test_raise_error_when_invalid_values_are_given(self, x: float) -> None:
        """Test invalid values."""
        with pytest.raises((LessOrEqualToZeroError, GreaterThan90Error)):
            cot(x)


class TestSec:
    """Validation for secant function."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        x = 60.0
        expected_result = 2.0
        assert sec(x) == pytest.approx(expected_result, rel=1e-4)

    @pytest.mark.parametrize("x", [-30.0])
    def test_raise_error_when_invalid_values_are_given(self, x: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, GreaterThan90Error)):
            sec(x)


class TestCsc:
    """Validation for cosecant function."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        x = 30.0
        expected_result = 2.0
        assert csc(x) == pytest.approx(expected_result, rel=1e-4)

    @pytest.mark.parametrize("x", [0.0, -30.0])
    def test_raise_error_when_invalid_values_are_given(self, x: float) -> None:
        """Test invalid values."""
        with pytest.raises((LessOrEqualToZeroError, GreaterThan90Error)):
            csc(x)



"""Testing formula 6.44 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_44 import Form6Dot44BetaRectangular
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot44BetaRectangular:
    """Validation for formula 6.44 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        u1 = 500.0
        u1_star = 400.0
        k = 1.2
        w_1 = 300.0
        e_par = 200.0

        # Object to test
        formula = Form6Dot44BetaRectangular(u1=u1, u1_star=u1_star, k=k, w_1=w_1, e_par=e_par)

        # Expected result, manually calculated
        manually_calculated_result = 401.25

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("u1", "u1_star", "k", "w_1", "e_par"),
        [
            (-500.0, 400.0, 1.2, 300.0, 200.0),  # u1 is negative
            (500.0, -400.0, 1.2, 300.0, 200.0),  # u1_star is negative
            (500.0, 400.0, -1.2, 300.0, 200.0),  # k is negative
            (500.0, 400.0, 1.2, -300.0, 200.0),  # w_1 is negative
            (500.0, 400.0, 1.2, 300.0, -200.0),  # e_par is negative
            (500.0, 0.0, 1.2, 300.0, 200.0),  # u1_star is zero
            (500.0, 400.0, 1.2, 0.0, 200.0),  # w_1 is zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, u1: float, u1_star: float, k: float, w_1: float, e_par: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot44BetaRectangular(u1=u1, u1_star=u1_star, k=k, w_1=w_1, e_par=e_par)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\beta = \frac{u_1}{u_{1^*}} + k \cdot \frac{u_1}{W_1} \cdot e_{par} = "
                r"\frac{500.000}{400.000} + 1.200 \cdot \frac{500.000}{300.000} \cdot 200.000 = 401.250 \ -",
            ),
            ("short", r"\beta = 401.250 \ -"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        u1 = 500.0
        u1_star = 400.0
        k = 1.2
        w_1 = 300.0
        e_par = 200.0

        # Object to test
        latex = Form6Dot44BetaRectangular(u1=u1, u1_star=u1_star, k=k, w_1=w_1, e_par=e_par).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Test for the stirrups configuration."""

import pytest
from shapely import Polygon

from blueprints.materials.reinforcement_steel import ReinforcementSteelMaterial
from blueprints.structural_sections.concrete.stirrups import StirrupConfiguration


class TestStirrup:
    """Tests for the stirrups configuration."""

    @pytest.fixture
    def stirrup(self) -> StirrupConfiguration:
        """Creates a stirrup configuration."""
        return StirrupConfiguration(
            geometry=Polygon([(0, 0), (0, 100), (100, 100), (100, 0)]),
            diameter=8,
            distance=100,
            material=ReinforcementSteelMaterial(),
        )

    def test_mandrel_diameter_factor(self, stirrup: StirrupConfiguration) -> None:
        """Test the mandrel diameter factor."""
        mandrel_diameter_factor = stirrup.mandrel_diameter_factor
        assert mandrel_diameter_factor == 4

    def test_mandrel_diameter_factor_none(self) -> None:
        """Test the mandrel diameter factor."""
        stirrup = StirrupConfiguration(
            geometry=Polygon([(0, 0), (0, 100), (100, 100), (100, 0)]),
            diameter=8,
            distance=100,
            material=ReinforcementSteelMaterial(),
            mandrel_diameter_factor=3.5,
        )
        assert stirrup.mandrel_diameter_factor == 3.5

    def test_as_w(self, stirrup: StirrupConfiguration) -> None:
        """Test total cross-sectional area of the stirrup [mm²/m]."""
        as_w = stirrup.as_w
        assert as_w == pytest.approx(expected=1005.3096, rel=1e-4)

    def test_area(self, stirrup: StirrupConfiguration) -> None:
        """Test the area of the stirrup."""
        area = stirrup.area
        assert area == pytest.approx(expected=50.2655, rel=1e-4)

    def test_radius(self, stirrup: StirrupConfiguration) -> None:
        """Test the radius of the stirrup."""
        radius = stirrup.radius
        assert radius == pytest.approx(expected=4, rel=1e-4)

    def test_centroid(self, stirrup: StirrupConfiguration) -> None:
        """Test the centroid of the stirrup."""
        centroid = stirrup.centroid
        assert centroid.x == pytest.approx(expected=50, rel=1e-4)
        assert centroid.y == pytest.approx(expected=50, rel=1e-4)

    def test_weight_per_meter(self, stirrup: StirrupConfiguration) -> None:
        """Test the weight per meter of the stirrup."""
        weight_per_meter = stirrup.weight_per_meter
        assert weight_per_meter == pytest.approx(expected=1.57833, rel=1e-4)

    def test_ctc_distance_legs(self, stirrup: StirrupConfiguration) -> None:
        """Test the distance between the legs of the stirrup."""
        ctc_distance_legs = stirrup.ctc_distance_legs
        assert ctc_distance_legs == pytest.approx(expected=100, rel=1e-4)

    def test_cover_used(self, stirrup: StirrupConfiguration) -> None:
        """Test the cover used."""
        assert stirrup.cover_used == 0

    def test_cover_used_changed(self) -> None:
        """Test the cover used."""
        stirrup = StirrupConfiguration(
            geometry=Polygon([(0, 0), (0, 100), (100, 100), (100, 0)]), diameter=8, distance=100, material=ReinforcementSteelMaterial(), cover_used=10
        )
        assert stirrup.cover_used == 10

    def test_relative_start_position(self, stirrup: StirrupConfiguration) -> None:
        """Test the relative start position."""
        assert stirrup.relative_start_position == 0.0

    def test_relative_end_position(self, stirrup: StirrupConfiguration) -> None:
        """Test the relative end position."""
        assert stirrup.relative_end_position == 1.0

    @pytest.mark.parametrize(("relative_start_position", "relative_end_position"), [(-1, 1), (-5, 1), (0, 2), (0, 5)])
    def test_relative_positions_wrong(self, relative_start_position: float, relative_end_position: float) -> None:
        """Test the relative positions."""
        with pytest.raises(ValueError):
            StirrupConfiguration(
                geometry=Polygon([(0, 0), (0, 100), (100, 100), (100, 0)]),
                diameter=8,
                distance=100,
                material=ReinforcementSteelMaterial(),
                relative_start_position=relative_start_position,
                relative_end_position=relative_end_position,
            )

    def test__repr__(self, stirrup: StirrupConfiguration) -> None:
        """Test the representation of the stirrup."""
        representation = repr(stirrup)
        assert representation.endswith("|⌀8/B500B")

    def test__str__(self, stirrup: StirrupConfiguration) -> None:
        """Test the string representation of the stirrup."""
        string_representation = str(stirrup)
        assert string_representation == "Stirrups ⌀8-100 mm | B500B | 1005.31 mm²/m"



"""Formula 6.4 from EN 1993-1-1:2005: Chapter 6 - Ultimate Limit State."""

from blueprints.codes.eurocode.en_1993_1_1_2005 import EN_1993_1_1_2005
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import MM, NMM, N
from blueprints.validations import raise_if_negative


class Form6Dot4AdditionalMoment(Formula):
    r"""Class representing formula 6.4 for the calculation of additional moment [$\Delta M_{Ed}$]."""

    label = "6.4"
    source_document = EN_1993_1_1_2005

    def __init__(
        self,
        n_ed: N,
        e_n: MM,
    ) -> None:
        r"""[$\Delta M_{Ed}$] Calculation of the additional moment [$Nmm$].

        EN 1993-1-1:2005 art.6.2.2.5(4) - Formula (6.4)
        Where a class 4 cross section is subjected to an axial compression force, the method given in EN 1993-1-5 should be used to
        determine the possible shift [$e_{N}$] of the centroid of the effective area [$A_{eff}$] relative to the centre of gravity
        of the gross cross section and the resulting additional moment according to this formula.

        Note: The sign of the additional moment depends on the effect in the combination of internal forces and moments, see 6.2.9.3(2).

        Parameters
        ----------
        n_ed : N
            [$N_{Ed}$] Axial compression force [$N$].
        e_n : MM
            [$e_{N}$] Shift of the centroid of the effective area relative to the centre of gravity of the gross cross section [$mm$].
            The method given in EN 1993-1-5 should be used to determine the possible shift [$e_{N}$] of the centroid of the effective
            area [$A_{eff$] relative to the centre of gravity of the gross cross section.
        """
        super().__init__()
        self.n_ed = n_ed
        self.e_n = e_n

    @staticmethod
    def _evaluate(
        n_ed: N,
        e_n: MM,
    ) -> NMM:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(n_ed=n_ed, e_n=e_n)

        return n_ed * e_n

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.4."""
        _equation: str = r"N_{Ed} \cdot e_{N}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"N_{Ed}": f"{self.n_ed:.{n}f}",
                r"e_{N}": f"{self.e_n:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"\Delta M_{Ed}",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="Nmm",
        )



"""Testing formula 5.29 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_5_structural_analysis.formula_5_29 import Form5Dot29BetaFactor
from blueprints.validations import LessOrEqualToZeroError


class TestForm5Dot29BetaFactor:
    """Validation for formula 5.29 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        c_0 = 8.0

        # Object to test
        formula = Form5Dot29BetaFactor(c_0=c_0)

        # Expected result, manually calculated
        manually_calculated_result = 1.2337  # dimensionless

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("c_0"),
        [
            (-8.0),  # c_0 is negative
            (0.0),  # c_0 is zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, c_0: float) -> None:
        """Test invalid values."""
        with pytest.raises(LessOrEqualToZeroError):
            Form5Dot29BetaFactor(c_0=c_0)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\beta = \frac{\pi^2}{c_0} = \frac{\pi^2}{8.000} = 1.234 \ -",
            ),
            ("short", r"\beta = 1.234 \ -"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        c_0 = 8.0

        # Object to test
        latex = Form5Dot29BetaFactor(c_0=c_0).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Common steel cross sectional shapes."""

from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.chs import CHS
from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.hea import HEA
from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.heb import HEB
from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.hem import HEM
from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.ipe import IPE
from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.rhs import RHS
from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.rhscf import RHSCF
from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.shs import SHS
from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.shscf import SHSCF
from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.strip import Strip
from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.unp import UNP

__all__ = [
    "CHS",
    "HEA",
    "HEB",
    "HEM",
    "IPE",
    "RHS",
    "RHSCF",
    "SHS",
    "SHSCF",
    "UNP",
    "Strip",
]



"""Testing formula 6.16 of EN 1993-1-1:2005."""

import pytest

from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_6_ultimate_limit_state.formula_6_16 import Form6Dot16CheckFlangeWithFastenerHoles
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot16CheckFlangeWithFastenerHoles:
    """Validation for formula 6.16 from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        a_f_net = 100.0
        f_u = 400.0
        gamma_m2 = 1.1
        a_f = 120.0
        f_y = 250.0
        gamma_m0 = 1.0

        # Object to test
        formula = Form6Dot16CheckFlangeWithFastenerHoles(a_f_net=a_f_net, f_u=f_u, gamma_m2=gamma_m2, a_f=a_f, f_y=f_y, gamma_m0=gamma_m0)

        # Expected result, manually calculated
        expected_result = True

        assert formula == expected_result

    @pytest.mark.parametrize(
        ("a_f_net", "f_u", "gamma_m2", "a_f", "f_y", "gamma_m0"),
        [
            (-100.0, 400.0, 1.1, 120.0, 250.0, 1.0),  # a_f_net is negative
            (100.0, -400.0, 1.1, 120.0, 250.0, 1.0),  # f_u is negative
            (100.0, 400.0, -1.1, 120.0, 250.0, 1.0),  # gamma_m2 is negative
            (100.0, 400.0, 1.1, -120.0, 250.0, 1.0),  # a_f is negative
            (100.0, 400.0, 1.1, 120.0, -250.0, 1.0),  # f_y is negative
            (100.0, 400.0, 1.1, 120.0, 250.0, -1.0),  # gamma_m0 is negative
            (100.0, 400.0, 0.0, 120.0, 250.0, 1.0),  # gamma_m2 is zero
            (100.0, 400.0, 1.1, 120.0, 250.0, 0.0),  # gamma_m0 is zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(
        self, a_f_net: float, f_u: float, gamma_m2: float, a_f: float, f_y: float, gamma_m0: float
    ) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot16CheckFlangeWithFastenerHoles(a_f_net=a_f_net, f_u=f_u, gamma_m2=gamma_m2, a_f=a_f, f_y=f_y, gamma_m0=gamma_m0)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"CHECK \to \left( \frac{A_{f,net} \cdot 0.9 \cdot f_{u}}{\gamma_{M2}} \geq \frac{A_{f} \cdot f_{y}}{\gamma_{M0}} \right) \to "
                r"\left( \frac{100.000 \cdot 0.9 \cdot 400.000}{1.100} \geq \frac{120.000 \cdot 250.000}{1.000} \right) \to OK",
            ),
            ("short", r"CHECK \to OK"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        a_f_net = 100.0
        f_u = 400.0
        gamma_m2 = 1.1
        a_f = 120.0
        f_y = 250.0
        gamma_m0 = 1.0

        # Object to test
        latex = Form6Dot16CheckFlangeWithFastenerHoles(a_f_net=a_f_net, f_u=f_u, gamma_m2=gamma_m2, a_f=a_f, f_y=f_y, gamma_m0=gamma_m0).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Formula 5.18 from EN 1993-5:2007: Chapter 5 - Ultimate limit state."""

from blueprints.codes.eurocode.en_1993_5_2007 import EN_1993_5_2007
from blueprints.codes.formula import ComparisonFormula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import KN
from blueprints.validations import raise_if_less_or_equal_to_zero


class Form5Dot18CompressionCheckUProfilesClass1And2(ComparisonFormula):
    r"""Class representing formula 5.18 for U-profiles of class 1 and 2: [$\frac{N_{Ed}}{N_{pl,Rd}} \leq 0.25$]."""

    label = "5.18"
    source_document = EN_1993_5_2007

    def __init__(
        self,
        n_ed: KN,
        n_pl_rd: KN,
    ) -> None:
        r"""Compression check for U-profiles of class 1 and 2.
        Design axial force [$N_{Ed}$] should not exceed 25% of plastic resistance [$N_{pl,Rd}$].

        EN 1993-5:2007 art. 5.2.3 (10) - Formula (5.18)

        Parameters
        ----------
        n_ed : KN
            [$N_{Ed}$] Design value of the compression force [$kN$].
        n_pl_rd : KN
            [$N_{pl,Rd}$] Plastic design resistance of the cross-section [$kN$].
        """
        super().__init__()
        self.n_ed = n_ed
        self.n_pl_rd = n_pl_rd

    @staticmethod
    def _evaluate_lhs(
        n_ed: KN,
        n_pl_rd: KN,
    ) -> float:
        """Evaluates the left-hand side of the comparison; see __init__ for details."""
        raise_if_less_or_equal_to_zero(n_pl_rd=n_pl_rd)
        return n_ed / n_pl_rd

    @staticmethod
    def _evaluate_rhs(*_, **_kwargs) -> float:
        """Evaluates the right-hand side of the comparison; see __init__ for details."""
        return 0.25

    @staticmethod
    def _evaluate(
        n_ed: KN,
        n_pl_rd: KN,
    ) -> bool:
        """Evaluates the comparison; see __init__ for details."""
        return (
            Form5Dot18CompressionCheckUProfilesClass1And2._evaluate_lhs(n_ed=n_ed, n_pl_rd=n_pl_rd)
            <= Form5Dot18CompressionCheckUProfilesClass1And2._evaluate_rhs()
        )

    @property
    def unity_check(self) -> float:
        """Returns the unity check value."""
        return self.lhs

    def __bool__(self) -> bool:
        """Allow truth-checking of the check object itself."""
        return self._evaluate(self.n_ed, self.n_pl_rd)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 5.18."""
        _equation: str = r"\frac{N_{Ed}}{N_{pl,Rd}} \leq 0.25"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"N_{Ed}": f"{self.n_ed:.{n}f}",
                r"N_{pl,Rd}": f"{self.n_pl_rd:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"CHECK",
            result="OK" if bool(self) else r"\text{Not OK}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label=r"\to",
            unit="",
        )



"""Testing sub formula 1 of formula 8.10 from EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_8_detailing_of_reinforcement_and_prestressing_tendons.formula_8_10 import (
    SubForm8Dot10Alpha6,
)
from blueprints.validations import NegativeValueError


class TestSubFormula8Dot8Alpha6:
    """Validation for sub formula 8.8 alpha 6 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # example values
        rho_1 = 32.5  # [-]
        sub_form_8_8_alpha_6 = SubForm8Dot10Alpha6(rho_1=rho_1)

        # manually calculated result
        manually_calculated_result = 1.140175  # [-]

        assert sub_form_8_8_alpha_6 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_evaluation_minimum(self) -> None:
        """Test the evaluation of the result if the minimum is reached."""
        # example values
        rho_1 = 62.5  # [-]
        sub_form_8_8_alpha_6 = SubForm8Dot10Alpha6(rho_1=rho_1)

        # manually calculated result
        manually_calculated_result = 1.5  # [-]

        assert sub_form_8_8_alpha_6 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_evaluation_maximum(self) -> None:
        """Test the evaluation of the result if the minimum is reached."""
        # example values
        rho_1 = 0.625  # [-]
        sub_form_8_8_alpha_6 = SubForm8Dot10Alpha6(rho_1=rho_1)

        # manually calculated result
        manually_calculated_result = 1  # [-]

        assert sub_form_8_8_alpha_6 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_evaluation_negative_rho_1(self) -> None:
        """Test the evaluation of the result if rho_1 is negative."""
        # example values
        rho_1 = -0.01  # [-]

        with pytest.raises(NegativeValueError):
            SubForm8Dot10Alpha6(rho_1=rho_1)

    @pytest.mark.parametrize(
        ("representation", "expected_result"),
        [
            (
                "complete",
                r"\alpha_6 = \max \left\{\min \left\{\left(\frac{\rho_1}{25}\right)^{0.5}; 1.5\right\}; 1\right\} = \max \left\{\min "
                r"\left\{\left(\frac{32.50}{25}\right)^{0.5}; 1.5\right\}; 1\right\} = 1.14",
            ),
            ("short", r"\alpha_6 = 1.14"),
        ],
    )
    def test_latex(self, representation: str, expected_result: str) -> None:
        """Test the LaTeX representation of the formula."""
        # example values
        rho_1 = 32.5

        latex = SubForm8Dot10Alpha6(rho_1=rho_1).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert actual[representation] == expected_result, f"{representation} representation failed."
